from typing import Iterable, List, Tuple

import numpy as np
import pytest
from brownie import Contract as BrownieContract
from brownie import chain
from brownie.convert.datatypes import EthAddress
from eth_utils.toolz import concat
from y import convert
from y.classes.common import ERC20
from y.contracts import Contract, contract_creation_block
from y.datatypes import Address, Block
from y.prices.dex.uniswap.v1 import UniswapV1

from y.networks import Network

mainnet_only = pytest.mark.skipif(
    chain.id != Network.Mainnet, reason="This test is only applicable on mainnet"
)
"""
A pytest marker to skip tests that are only applicable on the Ethereum mainnet.
"""

optimism_only = pytest.mark.skipif(
    chain.id != Network.Optimism, reason="This test is only applicable on optimism"
)
"""
A pytest marker to skip tests that are only applicable on the Optimism network.
"""

async_test = pytest.mark.asyncio_cooperative
"""
A pytest marker for concurrent asynchronous tests using asyncio_cooperative.
"""


def blocks_for_contract(address: Address, count: int = 5) -> List[Block]:
    """
    Generate a list of block numbers for testing a contract.

    This function generates a list of block numbers that are evenly spaced between the block
    obtained from :func:`~y.contracts.contract_creation_block` plus an offset of 10,000 and the current
    block height.

    Args:
        address: The address of the contract.
        count: Number of blocks to generate; default is 5.

    Returns:
        A list of block numbers where the first number is computed as
        contract_creation_block(address) + 10,000 and the remaining numbers are evenly spaced up to the current block height.

    Examples:
        Suppose that :func:`~y.contracts.contract_creation_block` returns 1000 and the current block height is 20000.
        Then the first block is 1000 + 10000 = 11000.
        With count=3, one possible output is:

        >>> blocks = blocks_for_contract("0x123", count=3)
        >>> blocks
        [11000, 15500, 20000]
        >>> isinstance(blocks, list)
        True

    See Also:
        :func:`~y.contracts.contract_creation_block`
    """
    address = convert.to_address(address)
    return [
        int(block)
        for block in np.linspace(contract_creation_block(address) + 10000, chain.height, count)
    ]


def mutate_address(address: Address) -> Tuple[str, str, str, EthAddress]:
    """
    Returns the same address in various forms for testing.

    This function returns a tuple with various representations of the given address:
    1. Lowercase string.
    2. Checksum address string.
    3. Address object produced via :func:`~y.convert.to_address`.
    4. EthAddress object.

    Args:
        address: The original address.

    Returns:
        A tuple containing:
            - Lowercase string.
            - Checksum address string.
            - Address object.
            - EthAddress object.

    Examples:
        >>> result = mutate_address("0xAbCdEf1234567890")
        >>> result[0]
        '0xabcdef1234567890'
        >>> isinstance(result[2], str)
        True

    See Also:
        :func:`~y.convert.to_address`
    """
    return (
        address.lower(),
        address[:2] + address[2:].upper(),
        convert.to_address(address),
        EthAddress(address),
    )


def mutate_addresses(addresses: Iterable[Address]):
    """
    Return a list of all mutated address representations for each address in the given iterable.

    Examples:
        >>> results = mutate_addresses(["0xAbCdEf1234567890", "0x1234567890abcdef"])
        >>> len(results)
        8

    See Also:
        :func:`mutate_address`
    """
    return list(concat(map(mutate_address, addresses)))


def mutate_contract(
    contract_address: Address,
) -> Tuple[str, str, str, EthAddress, BrownieContract]:
    """
    Returns the same contract address in various forms for testing.

    This function returns a tuple including various mutated representations:
    1-4. The different address forms returned by :func:`mutate_address`.
    5. A Brownie Contract object created with :class:`~brownie.Contract`.

    Args:
        contract_address: The original contract address.

    Returns:
        A tuple containing:
            - Lowercase string.
            - Checksum address string.
            - Address object.
            - EthAddress object.
            - Brownie Contract object.

    Examples:
        >>> result = mutate_contract("0xAbCdEf1234567890")
        >>> isinstance(result[4], BrownieContract)
        True

    See Also:
        :func:`mutate_address`
    """
    mutations = list(mutate_address(contract_address))
    mutations.append(Contract(contract_address))
    return tuple(mutations)


def mutate_contracts(addresses: Iterable[Address]):
    """
    Return a list of all mutated contract representations for each address in the given iterable.

    Each address is mutated into five representations (see :func:`mutate_contract`),
    so if two addresses are provided, the output will contain 10 items.

    Examples:
        >>> results = mutate_contracts(["0xAbCdEf1234567890", "0x1234567890abcdef"])
        >>> results  # Should contain 10 items
        [<mutated item>, <mutated item>, ..., <mutated item>]
        >>> len(results) == 10
        True

    See Also:
        :func:`mutate_contract`
    """
    return list(concat(map(mutate_contract, addresses)))


def mutate_token(
    token: Address,
) -> Tuple[str, str, str, EthAddress, BrownieContract, ERC20, int]:
    """
    Returns the same token address in various forms for testing.

    This function returns a tuple with different representations of the token address in the following order:
    1. Lowercase string.
    2. Checksum address string.
    3. Address object produced via :func:`~y.convert.to_address`.
    4. EthAddress object.
    5. Brownie Contract object created with :class:`~y.contracts.Contract`.
    6. An ERC20 object created with :class:`~y.classes.common.ERC20`.
    7. The integer conversion of the token address in hexadecimal format.

    Args:
        token: The original token address.

    Returns:
        A tuple containing:
            - Lowercase string representation.
            - Checksum string representation.
            - Address object.
            - EthAddress object.
            - Brownie Contract object.
            - ERC20 object.
            - Integer representation of the token address.

    Examples:
        >>> result = mutate_token("0xAbCdEf1234567890")
        >>> isinstance(result[5], ERC20)
        True
        >>> isinstance(result[6], int)
        True

    See Also:
        :func:`mutate_contract`
        :class:`~y.classes.common.ERC20`
    """
    mutations = list(mutate_contract(token))
    mutations.append(ERC20(token))
    mutations.append(int(token, 16))
    return tuple(mutations)


def mutate_tokens(addresses: Iterable[Address]):
    """
    Return a list of all mutated token representations for each address in the given iterable.

    Each address is mutated into seven representations (see :func:`mutate_token`),
    so if two addresses are provided, the output will contain 14 items.

    Examples:
        >>> results = mutate_tokens(["0xAbCdEf1234567890", "0x1234567890abcdef"])
        >>> len(results) == 14
        True

    See Also:
        :func:`mutate_token`
    """
    return list(concat(map(mutate_token, addresses)))


@pytest.fixture
def async_uni_v1():
    """
    A pytest fixture that yields an instance of :class:`~y.prices.dex.uniswap.v1.UniswapV1` in asynchronous mode.

    Examples:
        >>> async def test_example(async_uni_v1):
        ...     uniswap = async_uni_v1
        ...     # Use uniswap in asynchronous tests
        ...     assert uniswap.asynchronous is True

    See Also:
        :class:`~y.prices.dex.uniswap.v1.UniswapV1`
    """
    yield UniswapV1(asynchronous=True)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

from sphinx.util import logging


project = "ypricemagic"
copyright = "2024, BobTheBuidler"
author = "BobTheBuidler"

logger = logging.getLogger(__name__)

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "a_sync.sphinx.ext",
]

templates_path = ["_templates"]
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "../y/interfaces/*",
]

# Looks for objects in documentation of external libraries
intersphinx_mapping = {
    "a_sync": ("https://bobthebuidler.github.io/ez-a-sync", None),
    "brownie": ("https://eth-brownie.readthedocs.io/en/stable/", None),
    "checksum_dict": ("https://bobthebuidler.github.io/checksum_dict", None),
    "dank_mids": ("https://bobthebuidler.github.io/dank_mids", None),
    "dictstruct": ("https://bobthebuidler.github.io/dictstruct", None),
    "eth_typing": ("https://eth-typing.readthedocs.io/en/stable/", None),
    "evmspec": ("https://bobthebuidler.github.io/evmspec", None),
    "hexbytes": ("https://hexbytes.readthedocs.io/en/stable/", None),
    "pony": ("https://docs.ponyorm.org/", None),
    "python": ("https://docs.python.org/3", None),
    "typed_envs": ("https://bobthebuidler.github.io/typed-envs/", None),
    "typing_extensions": ("https://typing-extensions.readthedocs.io/en/latest/", None),
    "web3": ("https://web3py.readthedocs.io/en/stable/", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

autodoc_default_options = {
    "undoc-members": True,
    "special-members": True,
    "inherited-members": True,
    "member-order": "groupwise",
    # hide private methods that aren't relevant to us here
    "exclude-members": ",".join(
        [
            "__new__",
            "__dict__",
            "__slots__",
            "__module__",
            "__hash__",
            "__repr__",
            "__del__",
            "__setattr__",
            "__delattr__",
            "__init_from_abi__",
            "__method_names__",
            "__get_method_object__",
            "__annotations__",
            "__weakref__",
            "__reduce__",
            "__cause__",
            "__context__",
            "__suppress_context__",
            "__setstate__",
            "__traceback__",
            "_abc_impl",
            "__abstractmethods__",
            "__a_sync_default_mode__",
            "__a_sync_instance_will_be_sync__",
            "__a_sync_should_await__",
            "__a_sync_should_await_from_kwargs__",
            "__a_sync_flag_name__",
            "__a_sync_flag_value__",
            "__a_sync_instance_should_await__",
            "__a_sync_modifiers__",
            "__class_getitem__",
            "__init_subclass__",
            "__subclasshook__",
            "__orig_bases__",
            "__parameters__",
            # cache dict for instances for async_property lib
            "__async_property__",
            # float stuff
            "bit_count",
            "bit_length",
            "as_integer_ratio",
            "conjugate",
            "is_integer",
            "imag",
            "real",
        ]
    ),
}
autodoc_typehints = "description"
# Don't show class signature with the class' name.
autodoc_class_signature = "separated"

automodule_generate_module_stub = True

sys.path.insert(0, os.path.abspath("./y"))


def skip_specific_members(app, what, name, obj, skip, options):
    """
    Function to exclude specific members for a particular module.

    Args:
        app: The Sphinx application object.
        what: The type of the object which the docstring belongs to (one of "module", "class", "exception", "function", "method", "attribute").
        name: The fully qualified name of the object.
        obj: The object itself.
        skip: A boolean indicating if autodoc will skip this member if True.
        options: The options given to the directive: an object with attributes inherited_members, undoc_members, show_inheritance and noindex that are true if the flag option of same name was given to the auto directive.

    Returns:
        True if the member should be skipped, False otherwise.
    """

    # Skip some dundermethods in all cases.
    if name in [
        "__abs__",
        "__ceil__",
        "__divmod__",
        "__float__",
        "__floor__",
        "__format__",
        "__getformat__",
        "__getnewargs__",
        "__int__",
        "__mod__",
        "__ne__",
        "__neg__",
        "__pos__",
        "__rdivmod__",
        "__rfloordiv__",
        "__rmod__",
        "__rmul__",
        "__round__",
        "__rpow__",
        "__rsub__",
        "__rtruediv__",
        "__setformat__",
        "__trunc__",
        # not a dunder but this can go here
        "_ChecksumAddressSingletonMeta__instances",
    ]:
        return True

    # Skip the __init__ and __call__ members of any NewType objects we defined.
    if type(getattr(obj, "__self__", None)).__qualname__ == "typing.NewType" and name in [
        "__init__",
        "__call__",
    ]:
        return True

    # Skip the __init__, __str__, __getattribute__, args, and with_traceback members of all Exceptions
    if issubclass(getattr(obj, "__objclass__", type), BaseException) and name in [
        "__init__",
        "__str__",
        "__getattribute__",
        "args",
        "with_traceback",
    ]:
        return True

    if not skip:
        logger.info(f"module: {getattr(obj, '__module__', None)}  name: {name}  obj: {obj}")
    return skip


def setup(app):
    """
    Connect the skip_specific_members function to the autodoc-skip-member event.

    Args:
        app: The Sphinx application object.
    """
    app.connect("autodoc-skip-member", skip_specific_members)



from abc import abstractmethod
from asyncio import Task, create_task, ensure_future, get_event_loop
from decimal import Decimal
from functools import cached_property
from logging import getLogger
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Final,
    Generator,
    Literal,
    NoReturn,
    Optional,
    Tuple,
    Union,
    final,
)

import a_sync
from a_sync import cgather
from a_sync.a_sync import HiddenMethodDescriptor
from a_sync.a_sync.method import ASyncBoundMethod
from brownie import Contract, chain, web3
from brownie.convert.datatypes import HexString
from brownie.exceptions import ContractNotFound
from eth_retry import auto_retry
from eth_typing import ChecksumAddress
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y._decorators import stuck_coro_debugger
from y.classes.singleton import ChecksumASyncSingletonMeta
from y.constants import EEE_ADDRESS
from y.contracts import (
    Contract,
    build_name,
    contract_creation_block_async,
    has_method,
    probe,
)
from y.datatypes import Address, AnyAddressType, Block, Pool, UsdPrice
from y.exceptions import ContractNotVerified, MessedUpBrownieContract, NonStandardERC20
from y.networks import Network
from y.utils import _erc20, logging, raw_calls

if TYPE_CHECKING:
    from y.utils.events import Events

logger = getLogger(__name__)


def hex_to_string(h: HexString) -> str:
    """
    Convert a HexString to a string.

    Args:
        h: The HexString to convert.

    Returns:
        The converted string.
    """
    h = h.hex().rstrip("0")
    if len(h) % 2 != 0:
        h += "0"
    return bytes.fromhex(h).decode("utf-8")


class ContractBase(a_sync.ASyncGenericBase, metaclass=ChecksumASyncSingletonMeta):
    # defaults are stored as class vars to keep instance dicts smaller

    asynchronous: bool = False
    """A boolean indicating whether the instance will be used asynchronously."""

    _deploy_block: Optional[int] = None
    """The block when the contract was deployed, if known."""

    __slots__ = ("address",)

    def __init__(
        self,
        address: AnyAddressType,
        *,
        asynchronous: bool = False,
        _deploy_block: Optional[int] = None,
    ) -> None:
        self.address = convert.to_address(address)
        """
        The address of the contract.
        """

        if asynchronous:
            self.asynchronous = asynchronous
        if _deploy_block:
            self._deploy_block = _deploy_block
        super().__init__()

    def __str__(self) -> str:
        """
        Return the contract address as a string.

        Returns:
            The contract address as a string.
        """
        return f"{self.address}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} '{self.address}'>"

    def __eq__(self, __o: object) -> bool:
        # Skip checksumming if applicable, its computationally expensive
        if isinstance(__o, str):
            if self.address == __o:
                return True
            try:
                return self.address == convert.to_address(__o)
            except Exception:
                return False
        elif isinstance(__o, (ContractBase, Contract)):
            return self.address == __o.address
        return False

    def __hash__(self) -> int:
        return hash(self.address)

    @property
    def contract(self) -> Contract:
        return Contract(self.address)

    @cached_property
    def _is_cached(self) -> bool:
        try:
            self.contract
            return True
        except ContractNotVerified:
            return False
        except (ContractNotFound, MessedUpBrownieContract):
            return None

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def build_name(self) -> str:
        """
        Get the contract's build name.

        Returns:
            The contract's build name.

        Examples:
            >>> contract = ContractBase("0x1234567890abcdef1234567890abcdef12345678")
            >>> await contract.build_name
            'MyContract'
        """
        return await build_name(self.address, sync=False)

    __build_name__: HiddenMethodDescriptor[Self, str]

    @stuck_coro_debugger
    async def deploy_block(self, when_no_history_return_0: bool = False) -> int:
        """
        Get the block number when the contract was deployed.

        Args:
            when_no_history_return_0: If True, return 0 when no history is found instead of raising an exception.

        Returns:
            The block number when the contract was deployed, or 0 if `when_no_history_return_0` is True and the deploy block cannot be determined.

        Examples:
            >>> contract = ContractBase("0x1234567890abcdef1234567890abcdef12345678")
            >>> await contract.deploy_block()
            1234567

        See Also:
            - :func:`contract_creation_block_async`
        """
        if self._deploy_block is None:
            self._deploy_block = await contract_creation_block_async(
                self.address, when_no_history_return_0=when_no_history_return_0
            )
        return self._deploy_block

    deploy_block: ASyncBoundMethod[Self, Any, int]

    async def has_method(self, method: str, return_response: bool = False) -> Union[bool, Any]:
        """
        Check if the contract has a specific method.

        Args:
            method: The name of the method to check for.
            return_response: If True, return the response of the method call instead of a boolean.

        Returns:
            A boolean indicating whether the contract has the method, or the response of the method call if return_response is True.

        Examples:
            >>> contract = ContractBase("0x1234567890abcdef1234567890abcdef12345678")
            >>> await contract.has_method("name()")
            True

        See Also:
            - :func:`has_method`
        """
        return await has_method(self.address, method, return_response=return_response, sync=False)


class ERC20(ContractBase):
    """
    Represents an ERC20 token.
    """

    address: ChecksumAddress
    """
    The contract address of the token.
    """

    def __repr__(self) -> str:
        cls = type(self).__name__
        try:
            if ERC20.symbol.has_cache_value(self):
                symbol = ERC20.symbol.get_cache_value(self)
                return f"<{cls} {symbol} '{self.address}'>"
            elif not get_event_loop().is_running() and not self.asynchronous:
                try:
                    return f"<{cls} {self.__symbol__(sync=True)} '{self.address}'>"
                except NonStandardERC20:
                    return f"<{cls} SYMBOL_INVALID '{self.address}'>"
        except AttributeError:
            pass
        return f"<{cls} SYMBOL_NOT_LOADED '{self.address}'>"

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def symbol(self) -> str:
        """
        The token's symbol.

        Returns:
            The token's symbol.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token.symbol
            'TKN'
        """
        if self.address == EEE_ADDRESS:
            return {
                Network.Mainnet: "ETH",
                Network.Fantom: "FTM",
                Network.Polygon: "MATIC",
                Network.Arbitrum: "ETH",
                Network.Optimism: "ETH",
                Network.Base: "ETH",
                Network.Katana: "ETH",
                Network.Berachain: "BERA",
            }.get(chain.id, "ETH")
        import y._db.utils.token as db

        if symbol := await db.get_symbol(self.address):
            return symbol
        symbol = await self._symbol()
        db.set_symbol(self.address, symbol)
        return symbol

    @a_sync.aka.property
    @stuck_coro_debugger
    async def name(self) -> str:
        """
        The token's name.

        Returns:
            The token's name.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token.name
            'TokenName'
        """
        if self.address == EEE_ADDRESS:
            return "Ethereum"
        import y._db.utils.token as db

        name = await db.get_name(self.address)
        if name:
            return name
        name = await self._name()
        db.set_name(self.address, name)
        return name

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def decimals(self) -> int:
        """
        The number of decimal places for the token.

        Returns:
            The number of decimal places for the token.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token.decimals
            18
        """
        if self.address == EEE_ADDRESS:
            return 18
        import y._db.utils.token as db

        try:
            return await db.get_decimals(self.address)
        except ContractLogicError:
            # we've failed to fetch
            self.__raise_exception("decimals")
        except ValueError as e:
            if str(e).endswith(
                "of attr Token.decimals is greater than the maximum allowed value 2147483647"
            ):
                return await self._decimals()
            raise

    async def _decimals(self, block: Optional[Block] = None) -> int:
        """used to fetch decimals at specific block"""
        # TODO: deprecate and remove
        if self.address == EEE_ADDRESS:
            return 18
        try:
            return await _erc20.decimals(self.address, block=block, sync=False)
        except ContractLogicError:
            # we've failed to fetch
            self.__raise_exception("decimals")

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def scale(self) -> int:
        """
        Get the scaling factor for the token.

        Returns:
            The scaling factor for the token.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token.scale
            1000000000000000000
        """
        return 10 ** await self.__decimals__

    async def _scale(self, block: Optional[Block] = None) -> int:
        # TODO: deprecate and remove
        return 10 ** await self._decimals(block)

    async def total_supply(self, block: Optional[Block] = None) -> int:
        """
        Get the total supply of the token.

        Args:
            block (optional): The block number to query. Defaults to latest block.

        Returns:
            The total supply of the token.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token.total_supply()
            1000000000000000000000
        """
        return await _erc20.totalSupply(self.address, block=block, sync=False)

    async def total_supply_readable(self, block: Optional[Block] = None) -> float:
        """
        Get the total supply of the token scaled to a human-readable decimal.

        Args:
            block (optional): The block number to query.

        Returns:
            The total supply of the token scaled to a decimal.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token.total_supply_readable()
            1000.0
        """
        total_supply, scale = await cgather(
            self.total_supply(block=block, sync=False), self.__scale__
        )
        return total_supply / scale

    async def balance_of(self, address: AnyAddressType, block: Optional[Block] = None) -> int:
        """
        Query the balance of the token for a given address at a specific block.

        Args:
            address: The address to query.
            block (optional): The block number to query. Defaults to latest block.

        Returns:
            The balance of the token held by `address` at block `block`.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token.balance_of("0xabcdefabcdefabcdefabcdefabcdefabcdef")
            500000000000000000000
        """
        return await raw_calls.balanceOf(self.address, address, block=block, sync=False)

    async def balance_of_readable(
        self, address: AnyAddressType, block: Optional[Block] = None
    ) -> Decimal:
        balance, scale = await cgather(
            self.balance_of(address, block=block, sync=False), self.__scale__
        )
        return Decimal(balance) / scale

    async def price(
        self,
        block: Optional[Block] = None,
        return_None_on_failure: bool = False,
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: Tuple[Pool, ...] = (),
    ) -> Optional[UsdPrice]:
        """
        Get the price of the token in USD.

        Args:
            block (optional): The block number to query. Defaults to latest block.
            return_None_on_failure: If True, return None instead of raising a :class:`~y.exceptions.yPriceMagicError` on failure.
            skip_cache: If True, skip using the cache while fetching price data.
            ignore_pools: An optional tuple of pools to ignore when calculating the price.

        Returns:
            The price of the token in USD, or None if return_None_on_failure is True and the price cannot be retrieved.

        Raises:
            yPriceMagicError: If return_None_on_failure is False and the price cannot be retrieved.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token.price()
            1.23

        See Also:
            - :func:`y.prices.magic.get_price`
        """
        from y.prices.magic import get_price

        return await get_price(
            self.address,
            block=block,
            fail_to_None=return_None_on_failure,
            skip_cache=skip_cache,
            ignore_pools=ignore_pools,
            sync=False,
        )

    @classmethod
    async def _get_scale_for(cls, address: AnyAddressType) -> Awaitable[int]:
        # We use async ERC20 for memory sake since internal objects are
        # often async and we can avoid duplication 2x a/sync objs
        token = ERC20(address, asynchronous=True)
        try:
            return ERC20.scale.get_cache_value(token)
        except KeyError:
            # We'll use __scale__ instead of scale here for the purposes of optimization
            # We also pass sync kwarg to __scale__ to optimize speed even though async is default
            return await token.__scale__(sync=False)

    async def _symbol(self) -> str:
        """
        Get the token's symbol from the contract.

        Returns:
            The token's symbol.

        Raises:
            NonStandardERC20: If the symbol cannot be retrieved from the contract.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token._symbol()
            'TKN'
        """
        symbol = await probe(
            self.address,
            ("symbol()(string)", "SYMBOL()(string)", "getSymbol()(string)"),
        )
        if symbol is None:
            # Sometimes the above will fail if the symbol method returns bytes32, as with MKR. Let's try this.
            symbol = await probe(self.address, ("symbol()(bytes32)",))
            if symbol:
                symbol = hex_to_string(symbol)
        if symbol:
            return symbol
        # we've failed to fetch
        self.__raise_exception("symbol")

    async def _name(self) -> str:
        """
        Get the token's name from the contract.

        Returns:
            The token's name.

        Raises:
            NonStandardERC20: If the name cannot be retrieved from the contract.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> await token._name()
            'TokenName'
        """
        name = await probe(self.address, ("name()(string)", "NAME()(string)", "getName()(string)"))
        if name is None:
            # Sometimes the above will fail if the name method returns bytes32, as with MKR. Let's try this.
            name = await probe(self.address, ("name()(bytes32)",))
            if name:
                name = hex_to_string(name)
        if name:
            return name
        # we've failed to fetch
        self.__raise_exception("name")

    def __raise_exception(self, fn_name: str):
        """
        Raise a NonStandardERC20 exception with a custom error message.

        Args:
            fn_name: The name of the function that failed to retrieve data.

        Raises:
            NonStandardERC20: Always raised with a custom error message.

        Examples:
            >>> token = ERC20("0x1234567890abcdef1234567890abcdef12345678")
            >>> token.__raise_exception("symbol")
            Traceback (most recent call last):
            ...
            y.exceptions.NonStandardERC20: Unable to fetch `symbol` for 0x1234567890abcdef1234567890abcdef12345678 on Mainnet
        """
        raise NonStandardERC20(
            f"""
            Unable to fetch `{fn_name}` for {self.address} on {Network.printable()}
            If the contract is verified, please check to see if it has a strangely named
            `{fn_name}` method and create an issue on https://github.com/BobTheBuidler/ypricemagic
            with the contract address and correct method name so we can keep things going smoothly :)"""
        ) from None

    # These dundermethods are created by a_sync for the async_properties on this class
    __symbol__: HiddenMethodDescriptor[Self, str]
    __name__: HiddenMethodDescriptor[Self, str]
    __decimals__: HiddenMethodDescriptor[Self, int]
    __scale__: HiddenMethodDescriptor[Self, int]


@final
class WeiBalance(a_sync.ASyncGenericBase):
    """
    Represents a balance of a token in wei.

    This class provides utility methods for working with token balances in wei,
    including converting to readable format, getting the price in USD, and performing
    arithmetic operations.
    """

    # defaults are stored as class vars to keep instance dicts smaller
    block: Optional[Block] = None
    asynchronous: bool = False
    _skip_cache: bool = ENVS.SKIP_CACHE
    _ignore_pools: Tuple[Pool, ...] = ()
    __logger = None

    def __init__(
        self,
        balance: Union[int, Decimal],
        token: AnyAddressType,
        block: Optional[Block] = None,
        *,
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: Tuple[Pool, ...] = (),
        asynchronous: bool = False,
    ) -> None:
        """
        Initialize a WeiBalance object.

        Args:
            balance: The balance in wei.
            token: The token address.
            block: The block number for the balance (optional).
            skip_cache: If True, skip using the local cache for price data.
            ignore_pools: A tuple of pools to ignore when calculating the price.
            asynchronous: True if this object will be used with its asynchronous API, False for its sync API.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> print(balance)
            1000000000000000000
        """
        if asynchronous != self.asynchronous:
            self.asynchronous = asynchronous
        self.balance: Final = Decimal(balance)
        self.token: Final = ERC20(str(token), asynchronous=self.asynchronous)
        if block != self.block:
            self.block = block
        super().__init__()
        if skip_cache != self._skip_cache:
            self._skip_cache = skip_cache
        if ignore_pools != self._ignore_pools:
            self._ignore_pools = ignore_pools

    def __str__(self) -> str:
        """
        Return the balance in wei as a string.

        Returns:
            The balance in wei as a string.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> str(balance)
            '1000000000000000000'
        """
        return str(self.balance)

    def __repr__(self) -> str:
        return f"<WeiBalance token={self.token} balance={self.balance} block={self.block}>"

    def __hash__(self) -> int:
        return hash((self.balance, self.token, self.block, self._skip_cache, self._ignore_pools))

    def __bool__(self) -> bool:
        """
        Check if the balance is non-zero.

        Returns:
            True if the balance is non-zero, False otherwise.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> bool(balance)
            True
        """
        return bool(self.balance)

    def __eq__(self, __o: object) -> bool:
        """
        Check if two WeiBalance objects are equal.

        Args:
            __o: The object to compare with.

        Returns:
            True if the objects are equal, False otherwise.

        Examples:
            >>> balance1 = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> balance2 = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> balance1 == balance2
            True
        """
        if isinstance(__o, int):
            return __o == self.balance
        elif isinstance(__o, WeiBalance):
            return (
                self.balance == __o.balance
                and self.token == __o.token
                and self.block == __o.block
                and self._skip_cache == __o._skip_cache
                and self._ignore_pools == __o._ignore_pools
            )
        return False

    def __lt__(self, __o: object) -> bool:
        if isinstance(__o, int):
            return __o < self.balance
        elif isinstance(__o, WeiBalance):
            if self.token != __o.token:
                raise ValueError(
                    f"'<' only supported between {self.__class__.__name__} instances when denominated in the same token."
                ) from None
            return self.balance < __o.balance
        raise TypeError(
            f"'<' not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'"
        ) from None

    def __ge__(self, __o: object) -> bool:
        """
        Check if one WeiBalance object is greater than or equal to another.

        Args:
            __o: The object to compare with.

        Returns:
            True if the current object is greater than or equal to the other object, False otherwise.

        Raises:
            TypeError: If the objects are not of the same type.

        Examples:
            >>> balance1 = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> balance2 = WeiBalance(500000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> balance1 >= balance2
            True
        """
        if __o < self:
            return True
        elif type(__o) is type(self):
            return self == __o
        raise TypeError(
            f"'>=' not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'"
        ) from None

    def __radd__(self, __o: Union["WeiBalance", Literal[0]]) -> "WeiBalance":
        if __o == 0:
            return self
        try:
            if self.token != __o.token:
                raise ValueError(
                    f"addition not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'"
                ) from None
            if self.block != __o.block:
                raise ValueError(
                    "addition not supported between balances at different block heights"
                ) from None
            if self._skip_cache != __o._skip_cache:
                raise ValueError(
                    "addition not supported between balances with different `_skip_cache` values"
                ) from None
            if self._ignore_pools != __o._ignore_pools:
                raise ValueError(
                    "addition not supported between balances with different `_ignore_pools` values"
                ) from None
            return WeiBalance(
                self.balance - __o.balance,
                self.token,
                self.block,
                skip_cache=self._skip_cache,
                ignore_pools=self._ignore_pools,
            )
        except AttributeError:
            raise TypeError(
                f"right addition not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'"
            ) from None

    def __add__(self, __o: "WeiBalance") -> "WeiBalance":
        try:
            if self.token != __o.token:
                raise ValueError(
                    f"addition not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'"
                ) from None
            if self.block != __o.block:
                raise ValueError(
                    "addition not supported between balances at different block heights"
                ) from None
            if self._skip_cache != __o._skip_cache:
                raise ValueError(
                    "addition not supported between balances with different `_skip_cache` values"
                ) from None
            if self._ignore_pools != __o._ignore_pools:
                raise ValueError(
                    "addition not supported between balances with different `_ignore_pools` values"
                ) from None
            return WeiBalance(
                self.balance - __o.balance,
                self.token,
                self.block,
                skip_cache=self._skip_cache,
                ignore_pools=self._ignore_pools,
            )
        except AttributeError:
            raise TypeError(
                f"addition not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'"
            ) from None

    def __sub__(self, __o: "WeiBalance") -> "WeiBalance":
        try:
            if self.token != __o.token:
                raise ValueError(
                    f"subtraction not supported between instances of '{self.__class__.__name__}' and '{__o.__class__.__name__}'"
                ) from None
            if self.block != __o.block:
                raise ValueError(
                    "subtraction not supported between balances at different block heights"
                ) from None
            if self._skip_cache != __o._skip_cache:
                raise ValueError(
                    "subtraction not supported between balances with different `_skip_cache` values"
                ) from None
            if self._ignore_pools != __o._ignore_pools:
                raise ValueError(
                    "subtraction not supported between balances with different `_ignore_pools` values"
                ) from None
            return WeiBalance(
                self.balance - __o.balance,
                self.token,
                self.block,
                skip_cache=self._skip_cache,
                ignore_pools=self._ignore_pools,
            )
        except AttributeError:
            raise TypeError(
                f"subtraction not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'"
            ) from None

    def __mul__(self, __o: Union[int, float, Decimal]) -> "WeiBalance":
        """
        Multiply a WeiBalance object by a scalar value.

        Args:
            __o: The scalar value to multiply by.

        Returns:
            A new WeiBalance object representing the product.

        Raises:
            TypeError: If the scalar value is not a number.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> balance * 2
            <WeiBalance token=<ERC20 TKN '0x1234567890abcdef1234567890abcdef12345678'> balance=2000000000000000000 block=None>
        """
        if not isinstance(__o, (int, float, Decimal)):
            raise TypeError(
                f"multiplication not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'"
            ) from None
        return WeiBalance(
            self.balance * Decimal(__o),
            self.token,
            self.block,
            skip_cache=self._skip_cache,
            ignore_pools=self._ignore_pools,
        )

    def __truediv__(self, __o: Union[int, float, Decimal]) -> "WeiBalance":
        """
        Divide a WeiBalance object by a scalar value.

        Args:
            __o: The scalar value to divide by.

        Returns:
            A new WeiBalance object representing the quotient.

        Raises:
            TypeError: If the scalar value is not a number.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> balance / 2
            <WeiBalance token=<ERC20 TKN '0x1234567890abcdef1234567890abcdef12345678'> balance=500000000000000000 block=None>
        """
        if not isinstance(__o, (int, float, Decimal)):
            raise TypeError(
                f"division not supported between instances of '{type(self).__name__}' and '{type(__o).__name__}'"
            ) from None
        return WeiBalance(
            self.balance / Decimal(__o),
            self.token,
            self.block,
            skip_cache=self._skip_cache,
            ignore_pools=self._ignore_pools,
        )

    @a_sync.aka.property
    async def readable(self) -> Decimal:
        """
        Get the balance scaled to a human-readable decimal.

        Returns:
            The balance scaled to a decimal.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> await balance.readable
            Decimal('1.0')
        """
        if self.balance == 0:
            return 0
        scale = await self.token.__scale__
        readable = self.balance / scale
        self._logger.debug(
            "balance: %s  decimals: %s  readable: %s",
            self.balance,
            str(scale).count("0"),
            readable,
        )
        return readable

    __readable__: HiddenMethodDescriptor[Self, Decimal]

    @a_sync.aka.property
    async def price(self) -> Decimal:
        """
        Get the price of the token in USD.

        Returns:
            The price of the token in USD.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> await balance.price
            Decimal('1.23')
        """
        price = Decimal(
            await self.token.price(
                block=self.block,
                skip_cache=self._skip_cache,
                ignore_pools=self._ignore_pools,
                sync=False,
            )
        )
        self._logger.debug("balance: %s  price: %s", self, price)
        return price

    __price__: HiddenMethodDescriptor[Self, Decimal]

    @a_sync.aka.property
    async def value_usd(self) -> Decimal:
        """
        Get the value of the balance in USD.

        Returns:
            The value of the balance in USD.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> await balance.value_usd
            Decimal('1.23')
        """
        if self.balance == 0:
            return 0
        balance, price = await cgather(self.__readable__, self.__price__)
        value = balance * price
        self._logger.debug("balance: %s  price: %s  value: %s", balance, price, value)
        return value

    __value_usd__: HiddenMethodDescriptor[Self, Decimal]

    @property
    def _logger(self) -> logging.Logger:
        """
        Get the logger for the WeiBalance object.

        Returns:
            The logger for the WeiBalance object.

        Examples:
            >>> balance = WeiBalance(1000000000000000000, "0x1234567890abcdef1234567890abcdef12345678")
            >>> logger = balance._logger
        """
        logger = self.__logger
        if logger is None:
            logger = logging.get_price_logger(
                self.token.address, self.block, extra=self.__class__.__name__
            )
            self.__logger = logger
        return logger


class _Loader(ContractBase):
    """Used for use cases where you need to load data thru present time before proceeding, and then continue loading data in the background."""

    __slots__ = (
        "_loaded",
        "_init_block",
        "__exc",
        "__task",
    )

    def __init__(self, address: Address, *, asynchronous: bool = False):
        super().__init__(address, asynchronous=asynchronous)
        self._init_block = auto_retry(web3.eth.get_block_number)()
        self._loaded = None
        self.__exc = None
        self.__task = None

    def __await__(self) -> Generator[Any, None, Literal[True]]:
        """
        Returns `True` once the `_Loader` has loaded all relevant data thru the current block.

        Returns:
            A generator that yields `True` once the `_Loader` has loaded all relevant data thru the current block.

        Examples:
            >>> loader = _Loader("0x1234567890abcdef1234567890abcdef12345678")
            >>> await loader
            True
        """
        return self.loaded.__await__()

    @abstractmethod
    async def _load(self) -> NoReturn:
        """
        `self._load` is the coro that will run in the daemon task associated with this _Loader.
        Your implementation MUST set Event `self._loaded` once data has been loaded thru the current block, or it will hang indefinitely.
        """

    @property
    def loaded(self) -> Awaitable[Literal[True]]:
        """
        Returns `True` once the `_Loader` has loaded all relevant data thru the current block.

        Returns:
            An awaitable that resolves to `True` once the `_Loader` has loaded all relevant data thru the current block.

        Examples:
            >>> loader = _Loader("0x1234567890abcdef1234567890abcdef12345678")
            >>> await loader.loaded
            True
        """
        self._task  # ensure task is running and not errd
        if self._loaded is None:
            self._loaded = a_sync.Event(name=str(self))
        return self._loaded.wait()

    @property
    def _task(self) -> "Task[NoReturn]":
        """
        The task that runs `self._load()` for this `_Loader`.

        Returns:
            The task that runs `self._load()` for this `_Loader`.

        Raises:
            Exception: If the _Loader has any exception, it is raised.

        Examples:
            >>> loader = _Loader("0x1234567890abcdef1234567890abcdef12345678")
            >>> task = loader._task
        """
        if self.__exc:
            # create a new duplicate exc instead of building a massive traceback on the original
            raise type(self.__exc)(*self.__exc.args).with_traceback(self.__tb)
        if self.__task is None:
            logger.debug("creating loader task for %s", self)
            self.__task = create_task(coro=self.__load(), name=f"{self}.__load()")
            self.__task.add_done_callback(self._done_callback)
        return self.__task

    def _done_callback(self, task: "Task[Any]") -> None:
        """
        Called on `self._task` when it completes, if applicable.

        Args:
            task: The completed task.

        Examples:
            >>> loader = _Loader("0x1234567890abcdef1234567890abcdef12345678")
            >>> loader._done_callback(task)
        """
        if e := task.exception():
            logger.error("exception while loading %s: %s", self, e)
            logger.exception(e)
            self.__task = None

    async def __load(self) -> NoReturn:
        """
        Loads the loader and catches any exceptions.

        Examples:
            >>> loader = _Loader("0x1234567890abcdef1234567890abcdef12345678")
            >>> await loader.__load()
        """
        try:
            await self._load()
        except Exception as e:
            import traceback

            self.__exc = e
            self.__tb = e.__traceback__
            # no need to hold vars in memory
            traceback.clear_frames(self.__tb)
            raise


class _EventsLoader(_Loader):
    """
    Used for use cases where you need to load event data thru present time before proceeding,
    and then continue loading data in the background.
    """

    @property
    @abstractmethod
    def _events(self) -> "Events":
        """
        The Events object associated with this _EventsLoader.
        """

    @property
    def loaded(self) -> Awaitable[Literal[True]]:
        """
        Returns `True` once the `_Loader` has loaded all relevant data thru the current block.

        Returns:
            An awaitable that resolves to `True` once the `_Loader` has loaded all relevant data thru the current block.

        Examples:
            >>> loader = _EventsLoader("0x1234567890abcdef1234567890abcdef12345678")
            >>> await loader.loaded
            True
        """
        self._task  # ensure task is running and not err'd
        if self._loaded is None:
            self._loaded = ensure_future(self._events._lock.wait_for(self._init_block))
        return self._loaded

    async def _load(self) -> NoReturn:
        """
        Load the event data in the background.

        Examples:
            >>> loader = _EventsLoader("0x1234567890abcdef1234567890abcdef12345678")
            >>> await loader._load()
        """
        # TODO: extend this for constant loading
        async for _ in self._events.events(self._init_block):
            pass



from functools import lru_cache
from typing import Final, Optional, Set

import a_sync
import cchecksum
import hexbytes
from eth_typing import AnyAddress, ChecksumAddress, HexAddress

from y import ENVIRONMENT_VARIABLES as ENVS
from y.datatypes import AnyAddressType


HexBytes: Final = hexbytes.HexBytes

to_checksum_address: Final = cchecksum.to_checksum_address


@lru_cache(maxsize=ENVS.CHECKSUM_CACHE_MAXSIZE)  # type: ignore [call-overload]
def checksum(address: AnyAddress) -> ChecksumAddress:
    """Convert an address to its checksummed format.

    This function uses the `to_checksum_address` function to convert an Ethereum address
    to its checksummed format. If the address is invalid, a `ValueError` is raised.

    Args:
        address: The Ethereum address to be checksummed.

    Raises:
        ValueError: If the provided address is not a valid Ethereum address.

    Returns:
        The checksummed Ethereum address.

    Examples:
        >>> checksum("0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe")
        '0xDe0B295669a9FD93d5F28D9Ec85E40f4cb697BAe'

    See Also:
        - :func:`to_checksum_address` for the underlying implementation.
    """
    try:
        return to_checksum_address(address)
    except ValueError:
        raise ValueError(f"{repr(address)} is not a valid ETH address") from None


def to_address(address: AnyAddressType) -> ChecksumAddress:
    """Convert an address to a checksummed address, with caching.

    This function normalizes the input address to a string, checks if it is already
    checksummed and cached, and if not, computes the checksummed address and caches it.

    Args:
        address: The Ethereum address to be converted.

    Returns:
        The checksummed Ethereum address.

    Examples:
        >>> to_address("0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe")
        '0xDe0B295669a9FD93d5F28D9Ec85E40f4cb697BAe'

    See Also:
        - :func:`checksum` for the checksumming process.
        - :func:`__normalize_input_to_string` for input normalization.
    """
    address = __normalize_input_to_string(address)
    if checksummed := __get_checksum_from_cache(address):
        return checksummed
    checksummed = checksum(address)
    __cache_if_is_checksummed(address, checksummed)
    return checksummed  # type: ignore [return-value]


_checksum_thread: Final = a_sync.ThreadPoolExecutor(1)


async def to_address_async(address: AnyAddressType) -> ChecksumAddress:
    """Asynchronously convert an address to a checksummed address, with caching.

    This function normalizes the input address to a string, checks if it is already
    checksummed and cached, and if not, computes the checksummed address asynchronously
    and caches it.

    Args:
        address: The Ethereum address to be converted.

    Returns:
        The checksummed Ethereum address.

    Examples:
        >>> await to_address_async("0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe")
        '0xDe0B295669a9FD93d5F28D9Ec85E40f4cb697BAe'

    See Also:
        - :func:`checksum` for the checksumming process.
        - :func:`__normalize_input_to_string` for input normalization.
    """
    normalized = __normalize_input_to_string(address)
    if checksummed := __get_checksum_from_cache(normalized):
        return checksummed
    checksummed = await _checksum_thread.run(checksum, normalized)
    __cache_if_is_checksummed(normalized, checksummed)
    return checksummed


_is_checksummed: Final[Set[ChecksumAddress]] = set()
_is_not_checksummed: Final[Set[HexAddress]] = set()


def __get_checksum_from_cache(address: HexAddress) -> Optional[ChecksumAddress]:
    """Retrieve a checksummed address from the cache if available.

    This function checks if the given address is already known to be checksummed,
    and if so, returns the address. If the address is known to not be checksummed,
    it computes and returns the checksummed address.

    Args:
        address: The Ethereum address to check in the cache.

    Returns:
        The cached checksummed address if available, otherwise None.

    See Also:
        - :func:`checksum` for the checksumming process.
    """
    if address in _is_checksummed:
        return address  # type: ignore [return-value]
    elif address in _is_not_checksummed:
        # The checksum value is already in the lru-cache, there
        # is no reason to dispatch this function call to a thread.
        return checksum(address)
    else:
        return None


def __cache_if_is_checksummed(address: HexAddress, checksummed: ChecksumAddress) -> None:
    """Cache the address based on whether it is checksummed.

    This function adds the address to the `_is_checksummed` set if it is already
    checksummed, otherwise it adds it to the `_is_not_checksummed` set.

    Args:
        address: The original Ethereum address.
        checksummed: The checksummed Ethereum address.
    """
    if address == checksummed:
        _is_checksummed.add(address)
    else:
        _is_not_checksummed.add(address)


def __normalize_input_to_string(address: AnyAddressType) -> HexAddress:
    r"""Normalize the input address to a string format.

    This function converts the input address to a string format, handling different
    types such as strings, bytes, and integers. The function ensures that the address
    is in a consistent string format for further processing.

    Args:
        address: The Ethereum address to be normalized.

    Returns:
        The normalized address as a string.

    Examples:
        >>> __normalize_input_to_string("0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe")
        '0xde0B295669a9FD93d5F28D9Ec85E40f4cb697BAe'
        >>> __normalize_input_to_string(b'\xde\x0b)e\x66\x9a\x9f\xd9=_(\xd9\xec\x85\xe4\x0fL\xb6\x97\xba\xe')
        '0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae'
        >>> __normalize_input_to_string(123456789)
        '0x000000000000000000000000000000000000075b'

    See Also:
        - :class:`HexBytes` for handling byte conversions.
    """
    address_type = type(address)
    if address_type is str:
        return address  # type: ignore [return-value]
    elif issubclass(address_type, bytes):
        return (
            address.hex()  # type: ignore [union-attr, return-value]
            if address_type.__name__ == "HexBytes"
            else HexBytes(address).hex()
        )
    elif address_type is int:
        return _int_to_address(address)  # type: ignore [arg-type]
    else:
        return str(address)  # type: ignore [return-value]


def _int_to_address(int_address: int) -> HexAddress:
    """Convert an integer to an Ethereum address in hexadecimal format.

    This function converts an integer to a hexadecimal Ethereum address, ensuring
    it is properly padded to 40 characters.

    Args:
        int_address: The integer representation of the Ethereum address.

    Returns:
        The Ethereum address in hexadecimal format.

    Examples:
        >>> _int_to_address(123456789)
        '0x000000000000000000000000000000000000075b'
    """
    hex_value = HexBytes(int_address).hex()[2:]
    padding = "0" * (40 - len(hex_value))
    return f"0x{padding}{hex_value}"  # type: ignore [return-value]



import brownie.convert
import checksum_dict._utils
import dank_mids.brownie_patch.call
import eth_event.main
import eth_utils
import evmspec.data._main
import multicall.call
import web3._utils as web3_utils
import web3.main as web3_main
import web3.middleware as web3_middleware

import y.convert


def monkey_patch_checksum_cache():
    """
    Monkey patch dependency checksums with faster versions.

    This function replaces the default checksumming functions in various libraries
    with a faster implementation using `to_address`. This improves performance
    by utilizing caching and optimized checksumming.

    See Also:
        - :func:`to_address` for the checksumming process.
    """
    # this monkey patches brownie's convert.to_address with our lru_cache
    brownie.convert.to_address = y.convert.to_address

    # this monkey patches checksum_dict's checksumming with our lru_cache
    checksum_dict._utils.to_checksum_address = y.convert.to_address

    # this monkey patches the dank_mids brownie patch's checksums with our lru_cache
    dank_mids.brownie_patch.call.to_checksum_address = y.convert.to_address

    # this monkey patches eth_event's address checksumming with our lru_cache
    eth_event.main.to_checksum_address = y.convert.to_address

    # this monkey patches brownie's EthAddress class with faster execution
    eth_utils.to_checksum_address = y.convert.to_address

    # this monkey patches something else I don't remember now with faster execution
    eth_utils.address.to_checksum_address = y.convert.to_address

    # this monkey patches evmspec's Address decode hook with our lru cache
    evmspec.data._main.to_checksum_address = y.convert.to_address

    # this monkey patches multicall.Call.target checksumming with our lru cache
    multicall.call.to_checksum_address = y.convert.to_address

    # this monkey patches all checksums in web3py with faster execution
    web3_main.to_checksum_address = y.convert.to_address
    web3_utils.ens.to_checksum_address = y.convert.to_address
    web3_utils.method_formatters.to_checksum_address = y.convert.to_address
    web3_utils.normalizers.to_checksum_address = y.convert.to_address
    web3_middleware.signing.to_checksum_address = y.convert.to_address

    try:
        import web3.utils.address as web3_address

        web3_address.to_checksum_address = y.convert.to_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass

    try:
        import ens.ens  # type: ignore [import-untyped]

        ens.ens.to_checksum_address = y.convert.to_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass

    try:
        import ens.async_ens  # type: ignore [import-untyped]

        ens.async_ens.to_checksum_address = y.convert.to_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass



import logging
from typing import Optional

import a_sync
from brownie import chain

from y import convert
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

POOLS = {
    Network.BinanceSmartChain: {
        "0x86aFa7ff694Ab8C985b79733745662760e454169": "0xF16D312d119c13dD27fD0dC814b0bCdcaAa62dfD",  # Belt.fi bDAI/bUSDC/bUSDT/bBUSD
        "0x9cb73F20164e399958261c289Eb5F9846f4D1404": "0xAEA4f7dcd172997947809CE6F12018a6D5c1E8b6",  # 4Belt
    },
}.get(chain.id, {})


def is_belt_lp(token: AnyAddressType) -> bool:
    """Check if a token is a Belt LP token.

    Args:
        token: The address of the token to check.

    Returns:
        True if the token is a Belt LP token, False otherwise.

    Examples:
        >>> is_belt_lp("0x86aFa7ff694Ab8C985b79733745662760e454169")
        True
        >>> is_belt_lp("0x0000000000000000000000000000000000000000")
        False

    See Also:
        :data:`POOLS` for the list of recognized Belt LP tokens.
    """
    address = convert.to_address(token)
    return address in POOLS


@a_sync.a_sync(default="sync")
async def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    """Get the price of a Belt LP token in USD.

    This function retrieves the virtual price of a Belt LP token from its associated pool
    and converts it to a USD price.

    Args:
        token: The address of the Belt LP token.
        block: The block number at which to fetch the price. Defaults to the latest block.

    Raises:
        KeyError: If the token is not found in the :data:`POOLS` dictionary.

    Examples:
        >>> await get_price("0x86aFa7ff694Ab8C985b79733745662760e454169")
        UsdPrice(1.23456789)
        >>> await get_price("0x9cb73F20164e399958261c289Eb5F9846f4D1404", block=12345678)
        UsdPrice(1.23456789)

    See Also:
        :func:`is_belt_lp` to check if a token is a Belt LP token.
    """
    address = await convert.to_address_async(token)
    pool = POOLS[address]
    virtual_price = await raw_call(
        pool, "get_virtual_price()", output="int", block=block, sync=False
    )
    return UsdPrice(virtual_price / 10**18)



from decimal import Decimal
from logging import DEBUG, getLogger
from typing import Optional, Tuple

import a_sync
from a_sync import ASyncCachedPropertyDescriptor, cgather
from a_sync.a_sync import HiddenMethodDescriptor
from dank_mids.exceptions import Revert
from multicall.call import Call
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import Network
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20
from y.constants import CHAINID, CONNECTED_TO_MAINNET, weth
from y.contracts import Contract, has_method, has_methods, probe
from y.datatypes import AnyAddressType, Block, Pool, UsdPrice
from y.exceptions import (
    CantFetchParam,
    ContractNotVerified,
    MessedUpBrownieContract,
    PriceError,
    yPriceMagicError,
)
from y.utils.cache import optional_async_diskcache
from y.utils.logging import get_price_logger
from y.utils.raw_calls import raw_call

logger = getLogger(__name__)

# NOTE: Yearn and Yearn-inspired

underlying_methods = (
    "token()(address)",
    "underlying()(address)",
    "native()(address)",
    "want()(address)",
    "input()(address)",
    "asset()(address)",
    "wmatic()(address)",
    "wbnb()(address)",
    "based()(address)",
)
"""
List of methods which might be used to get the underlying asset of a vault.
"""

share_price_methods = (
    "pricePerShare()(uint)",
    "getPricePerShare()(uint)",
    "getPricePerFullShare()(uint)",
    "getSharesToUnderlying()(uint)",
    "exchangeRate()(uint)",
)
"""
List of methods which might be used to get the share price of a vault.
"""

# wibbtc returns True here even though it doesn't meet the criteria.
# TODO figure out a better fix. For now I need a fix asap so this works.
# NOTE turns out I never fixed it and there are more now. Womp.
force_false = {
    Network.Mainnet: (
        "0x8751D4196027d4e6DA63716fA7786B5174F04C15",  # wibBTC
        "0xF0a93d4994B3d98Fb5e3A2F90dBc2d69073Cb86b",  # PWRD
        "0x35Ec69A77B79c255e5d47D5A3BdbEFEfE342630c",  # ynLSDe
    ),
}


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=30 * 60)
@stuck_coro_debugger
@optional_async_diskcache
async def is_yearn_vault(token: AnyAddressType) -> bool:
    if str(token) in force_false.get(CHAINID, ()):
        return False

    # Yearn-like contracts can use these formats
    result = any(
        await cgather(
            has_methods(
                token,
                (
                    "pricePerShare()(uint)",
                    "getPricePerShare()(uint)",
                    "getPricePerFullShare()(uint)",
                    "getSharesToUnderlying()(uint)",
                ),
                any,
                sync=False,
            ),
            has_methods(token, ("exchangeRate()(uint)", "underlying()(address)"), sync=False),
        )
    )

    # pricePerShare can revert if totalSupply == 0, which would cause `has_methods` to return `False`,
    # but it might still be a vault. This section will correct `result` for problematic vaults.
    if not result:
        try:
            contract = await Contract.coroutine(token)
        except (ContractNotVerified, MessedUpBrownieContract):
            pass
        else:
            result = (
                hasattr(contract, "pricePerShare")
                or hasattr(contract, "getPricePerShare")
                or hasattr(contract, "getPricePerFullShare")
                or hasattr(contract, "getSharesToUnderlying")
                or hasattr(contract, "convertToAssets")
            )
    return result


@a_sync.a_sync(default="sync")
async def get_price(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
    ignore_pools: Tuple[Pool, ...] = (),
) -> UsdPrice:
    return await YearnInspiredVault(token).price(
        block=block, skip_cache=skip_cache, ignore_pools=ignore_pools, sync=False
    )


_BEEFY_METHODS = {
    "BeefyVaultV6Matic": "wmatic()",
    "BeefyVenusVaultBNB": "wbnb()",
}


class YearnInspiredVault(ERC20):
    """
    Represents a vault token from Yearn or a similar protocol.

    This class extends ERC20 and provides methods to interact with vaults,
    including fetching the underlying asset, share price, and token price.
    """

    # defaults are stored as class vars to keep instance dicts smaller

    _get_share_price = None
    """
    Cached method to get the share price.
    
    This class will probe various share price methods to find the correct one, and then save it for reuse.
    """
    # v1 vaults use getPricePerFullShare scaled to 18 decimals
    # v2 vaults use pricePerShare scaled to underlying token decimals
    # yearnish clones use all sorts of other things, we gotchu covered

    # mypy helpers
    underlying: ASyncCachedPropertyDescriptor[Self, ERC20]
    __underlying__: HiddenMethodDescriptor[Self, ERC20]

    @a_sync.aka.cached_property
    async def underlying(self) -> ERC20:
        """
        Fetches the underlying asset of the vault.

        Returns:
            The underlying ERC20 token.

        Raises:
            CantFetchParam: If the underlying asset cannot be determined.

        Special Cases:
            1. Arbitrum USDL: For the specific address 0x57c7E0D43C05bCe429ce030132Ca40F6FA5839d7 on Arbitrum,
               it uses the 'usdl()' method to fetch the underlying token.
            2. Beefy Vaults: For certain Beefy vaults (BeefyVaultV6Matic and BeefyVenusVaultBNB),
               it uses 'wmatic()' or 'wbnb()' methods respectively.
            3. Reaper Vaults: For certain Reaper vaults, it checks for a 'lendPlatform()' method
               and then queries the 'underlying()' method on the lend platform contract.

        Example:
            >>> vault = YearnInspiredVault("0x5f18C75AbDAe578b483E5F43f12a39cF75b973a9")  # yvUSDC
            >>> underlying = vault.underlying()
            >>> underlying.address
            "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"  # USDC address
        """
        # special cases
        if CHAINID == Network.Arbitrum:
            if self.address == "0x57c7E0D43C05bCe429ce030132Ca40F6FA5839d7":
                underlying = await raw_call(self.address, "usdl()", output="address", sync=False)
                return ERC20(underlying, asynchronous=self.asynchronous)
        elif CONNECTED_TO_MAINNET and self.address == "0x09db87A538BD693E9d08544577d5cCfAA6373A48":
            # ynETH yield nest has no method for underlying
            return ERC20(weth, asynchronous=self.asynchronous)

        try:
            underlying = await probe(self.address, underlying_methods)
        except AssertionError:
            # special handler for some strange beefy vaults
            beefy_method = _BEEFY_METHODS.get(await self.__build_name__)
            if beefy_method is None:
                raise
            underlying = await raw_call(self.address, beefy_method, output="address", sync=False)

        if underlying is None:
            # certain reaper vaults
            try:
                lend_platform = await self.has_method(
                    "lendPlatform()(address)", return_response=True, sync=False
                )
            except ContractLogicError:
                pass
            else:
                if lend_platform:
                    underlying = await has_method(
                        lend_platform,
                        "underlying()(address)",
                        return_response=True,
                        sync=False,
                    )

        if underlying is not None:
            underlying = ERC20(underlying, asynchronous=self.asynchronous)
            if logger.isEnabledFor(DEBUG):
                logger._log(
                    DEBUG,
                    "%s %s underlying is %s %s",
                    (
                        await self.__symbol__,
                        self,
                        await underlying.__symbol__,
                        underlying,
                    ),
                )
            return underlying
        raise CantFetchParam(f"underlying for {self}")

    @a_sync.a_sync(cache_type="memory", ram_cache_maxsize=1000)
    async def share_price(self, block: Optional[Block] = None) -> Optional[Decimal]:
        """
        Calculates the share price of the vault.

        Args:
            block (optional): The block number to query. Defaults to the latest block.

        Returns:
            The share price of the vault, or None if the vault's total supply is zero.

        Raises:
            CantFetchParam: If the share price cannot be fetched or calculated.

        Example:
            >>> vault = YearnInspiredVault("0x5f18C75AbDAe578b483E5F43f12a39cF75b973a9")  # yvUSDC
            >>> share_price = await vault.share_price(block=14_000_000)
            >>> print(f"{share_price:.6f}")
            1.096431
        """
        if self._get_share_price:
            try:
                share_price = await self._get_share_price.coroutine(block_id=block)
            except Exception as e:
                logger.debug("exc %s when fetching share price for %s", e, self)
                share_price = await probe(self.address, share_price_methods, block=block)
        else:
            share_price_method, share_price = await probe(
                self.address, share_price_methods, block=block, return_method=True
            )
            if share_price_method:
                self._get_share_price = Call(self.address, [share_price_method])

        if share_price is None:
            # this is for element vaults and other 'scaled' share price functions. probe fails because method requires input
            try:
                contract = await Contract.coroutine(self.address)
            except ContractNotVerified:
                pass
            else:
                for method in ("convertToAssets", "getSharesToUnderlying"):
                    if contract_call := getattr(contract, method, None):

                        scale = await self.__scale__

                        class call:
                            # a hacky way we can cache this weird case and save calls
                            function = method

                            @staticmethod
                            async def coroutine(block_id: Optional[Block]) -> int:
                                return await contract_call.coroutine(
                                    scale, block_identifier=block_id
                                )

                        try:
                            share_price = await call.coroutine(block_id=block)
                        except (ContractLogicError, Revert):
                            pass
                        else:
                            self._get_share_price = call

        if share_price is not None:
            if (
                self._get_share_price
                and self._get_share_price.function == "getPricePerFullShare()(uint)"
            ):
                # v1 vaults use getPricePerFullShare scaled to 18 decimals
                return share_price / Decimal(10**18)
            underlying = await self.__underlying__
            return Decimal(share_price) / await underlying.__scale__

        elif (
            await raw_call(
                self.address,
                "totalSupply()",
                output="int",
                block=block,
                return_None_on_failure=True,
                sync=False,
            )
            == 0
        ):
            return None

        else:
            raise CantFetchParam(f"share_price for {self}")

    a_sync.a_sync(cache_type="memory", ram_cache_maxsize=1000)

    async def price(
        self,
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> UsdPrice:
        """
        Calculates the USD price of the vault token.

        Args:
            block (optional): The block number to query. Defaults to the latest block.
            ignore_pools: Pools to ignore when calculating the price.
            skip_cache: Whether to skip the cache when fetching prices.

        Returns:
            The USD price of the vault token.

        Example:
            >>> vault = YearnInspiredVault("0x5f18C75AbDAe578b483E5F43f12a39cF75b973a9")  # yvUSDC
            >>> price = vault.price(block=14_000_000)
            >>> print(f"{price:.6f}")
            1.096431  # The price of yvUSDC in USD
        """
        logger = get_price_logger(self.address, block=None, extra="yearn")
        underlying: ERC20
        try:
            share_price, underlying = await cgather(
                self.share_price(block=block, sync=False), self.__underlying__
            )
        except CantFetchParam:
            return None
        if share_price is None:
            return None
        logger.debug("%s share price at block %s: %s", self, block, share_price)
        try:
            price = UsdPrice(
                share_price
                * Decimal(
                    await underlying.price(
                        block=block,
                        ignore_pools=ignore_pools,
                        skip_cache=skip_cache,
                        sync=False,
                    )
                )
            )
        except yPriceMagicError as e:
            if not isinstance(e.exception, PriceError):
                raise
            price = None
        logger.debug("%s price at block %s: %s", self, block, price)
        return price



from asyncio import Task, create_task
from contextlib import suppress
from enum import IntEnum
from logging import DEBUG, getLogger
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    List,
    NewType,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import a_sync
from a_sync import cgather
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import ZERO_ADDRESS
from brownie.convert.datatypes import EthAddress
from brownie.network.event import _EventItem
from faster_eth_abi.exceptions import InvalidPointer
from hexbytes import HexBytes
from multicall import Call
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import constants, contracts
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20, ContractBase, WeiBalance
from y.constants import CHAINID, CONNECTED_TO_MAINNET
from y.contracts import Contract
from y.datatypes import Address, AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import ContractNotVerified, TokenNotFound
from y.networks import Network
from y.prices.dex.balancer._abc import BalancerABC, BalancerPool
from y.utils.cache import a_sync_ttl_cache
from y.utils.events import ProcessedEvents
from y.utils.logging import get_price_logger


BALANCER_V2_VAULTS = {
    Network.Mainnet: [
        "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
    ],
    Network.Fantom: [
        "0x20dd72Ed959b6147912C2e529F0a0C651c33c9ce",
    ],
    Network.Polygon: [
        "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
    ],
    Network.Arbitrum: [
        "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
    ],
    Network.Base: [
        "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
    ],
    Network.Berachain: [
        "0x4Be03f781C497A489E3cB0287833452cA9B9E80B",  # BEX Exchange https://docs.bex.berachain.com/developers/
    ],
}.get(CHAINID, [])

MESSED_UP_POOLS = {
    Network.Mainnet: [
        # NOTE: this was the first ever balancer "pool" and isn't actually a pool
        "0xF3799CBAe9c028c1A590F4463dFF039b8F4986BE",
    ],
}.get(CHAINID, [])

T = TypeVar("T")

PoolId = NewType("PoolId", bytes)
PoolBalances = Dict[ERC20, WeiBalance]

logger = getLogger(__name__)


class PoolSpecialization(IntEnum):
    ComposableStablePool = 0
    WeightedPool = 1
    WeightedPool2Tokens = 2
    # This is a weird one
    CronV1Pool = -1

    @staticmethod
    def with_immutable_tokens() -> List["PoolSpecialization"]:
        """
        Get a list of pool specializations with immutable tokens.

        Returns:
            A list of :class:`~PoolSpecialization` enums representing pools with immutable tokens.

        Examples:
            >>> PoolSpecialization.with_immutable_tokens()
            [<PoolSpecialization.ComposableStablePool: 0>, <PoolSpecialization.WeightedPool: 1>, <PoolSpecialization.WeightedPool2Tokens: 2>, <PoolSpecialization.CronV1Pool: -1>]
        """
        return [
            PoolSpecialization.ComposableStablePool,
            PoolSpecialization.WeightedPool,
            PoolSpecialization.WeightedPool2Tokens,
            PoolSpecialization.CronV1Pool,
        ]


class BalancerV2Vault(ContractBase):
    def __init__(self, address: AnyAddressType, *, asynchronous: bool = False) -> None:
        """
        Initialize a BalancerV2Vault instance.

        Args:
            address: The address of the Balancer V2 Vault.
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> vault = BalancerV2Vault("0xBA12222222228d8Ba445958a75a0704d566BF2C8")
        """
        super().__init__(address, asynchronous=asynchronous)
        self._events = BalancerEvents(
            self,
            addresses=address,
            topics=["0x3c13bc30b8e878c53fd2a36b679409c073afd75950be43d8858768e956fbc20e"],
        )
        if not self._is_cached:
            # we need the contract cached so we can decode logs correctly
            self.contract

    @stuck_coro_debugger
    async def pools(self, block: Optional[Block] = None) -> AsyncIterator["BalancerV2Pool"]:
        """
        Asynchronously iterate over Balancer V2 pools.

        Args:
            block: The block number to query. Defaults to the latest block.

        Yields:
            Instances of :class:`~BalancerV2Pool`.

        Examples:
            >>> async for pool in vault.pools():
            ...     print(pool)
        """
        async for pool in self._events.events(to_block=block):
            yield pool

    @stuck_coro_debugger
    async def pools_for_token(
        self, token: Address, block: Optional[Block] = None
    ) -> AsyncIterator["BalancerV2Pool"]:
        """
        Asynchronously iterate over Balancer V2 pools containing a specific token.

        Args:
            token: The address of the token to search for.
            block: The block number to query. Defaults to the latest block.

        Yields:
            Instances of :class:`~BalancerV2Pool` containing the specified token.

        Examples:
            >>> async for pool in vault.pools_for_token("0xTokenAddress"):
            ...     print(pool)
        """
        tasks = a_sync.map(BalancerV2Pool.tokens, block=block)
        debug_logs = logger.isEnabledFor(DEBUG)
        async for pool in self.pools(block=block):
            if tokens := pool._tokens:
                if token in tokens:
                    if debug_logs:
                        logger._log(DEBUG, "%s contains %s", (pool, token))
                    yield pool
            else:
                # start the task now, we can await it later
                tasks[pool]

        if tasks:
            async for pool, tokens in tasks.map(pop=True):
                if token in tokens:
                    if debug_logs:
                        logger._log(DEBUG, "%s contains %s", (pool, token))
                    yield pool

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_pool_tokens(self, pool_id: HexBytes, block: Optional[Block] = None):
        """
        Get the tokens and balances for a specific pool.

        Args:
            pool_id: The ID of the pool.
            block: The block number to query. Defaults to the latest block.

        Returns:
            A tuple containing the tokens and their balances.

        Examples:
            >>> tokens, balances = await vault.get_pool_tokens(pool_id)
        """
        contract = await contracts.Contract.coroutine(self.address)
        try:
            return await contract.getPoolTokens.coroutine(pool_id, block_identifier=block)
        except InvalidPointer:
            return (), (), None

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_pool_info(
        self, poolids: Tuple[HexBytes, ...], block: Optional[Block] = None
    ) -> List[Tuple]:
        """
        Get information for multiple pools.

        Args:
            poolids: A tuple of pool IDs.
            block: The block number to query. Defaults to the latest block.

        Returns:
            A list of tuples containing pool information.

        Examples:
            >>> pool_info = await vault.get_pool_info((pool_id1, pool_id2))
        """
        contract = await contracts.Contract.coroutine(self.address)
        return await contract.getPoolTokens.map(poolids, block_identifier=block)

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def deepest_pool_for(
        self, token_address: Address, block: Optional[Block] = None
    ) -> "BalancerV2Pool":
        """
        Find the deepest pool for a specific token.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.

        Returns:
            The :class:`~BalancerV2Pool` with the deepest liquidity for the specified token.

        Examples:
            >>> deepest_pool = await vault.deepest_pool_for("0xTokenAddress")
        """
        balance_tasks: a_sync.TaskMapping[BalancerV2Pool, Optional[WeiBalance]]

        logger = get_price_logger(token_address, block, extra="balancer.v2")

        balance_tasks = BalancerV2Pool.get_balance.map(token_address=token_address, block=block)
        balances_aiterator = balance_tasks.map(
            self.pools_for_token(token_address, block=block), pop=True
        )
        filtered = balances_aiterator.filter(_lookup_balance_from_tuple)
        high_to_low = filtered.sort(key=_lookup_balance_from_tuple, reverse=True)

        if logger.isEnabledFor(DEBUG):
            async for pool, balance in high_to_low:
                logger._log(DEBUG, "deepest pool %s balance %s", (pool, balance))
                return pool
        else:
            async for pool, balance in high_to_low:
                return pool


class BalancerEvents(ProcessedEvents[Tuple[HexBytes, EthAddress, Block]]):
    __slots__ = ("asynchronous",)

    def __init__(self, vault: BalancerV2Vault, *args, asynchronous: bool = False, **kwargs):
        """
        Initialize a BalancerEvents instance.

        Args:
            vault: The associated :class:`~BalancerV2Vault`.
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> events = BalancerEvents(vault)
        """
        super().__init__(*args, **kwargs)
        self.vault = vault
        self.asynchronous = asynchronous
        self.__tasks = []

    def _include_event(self, event: _EventItem) -> Awaitable[bool]:
        """
        Determine whether to include a specific event.

        Args:
            event: The event to evaluate.

        Returns:
            A boolean indicating whether to include the event.

        Examples:
            >>> include = await events._include_event(event)
        """
        if event["poolAddress"] in MESSED_UP_POOLS:
            return False
        # NOTE: For some reason the Balancer fork on Fantom lists "0x3e522051A9B1958Aa1e828AC24Afba4a551DF37d"
        #       as a pool, but it is not a contract. This handler will prevent it and future cases from causing problems.
        # NOTE: this isn't really optimized as it still runs semi-synchronously but its better than what was had previously
        return self.executor.run(contracts.is_contract, event["poolAddress"])

    def _process_event(self, event: _EventItem) -> "BalancerV2Pool":
        """
        Process a specific event and return the associated Balancer V2 pool.

        Args:
            event: The event to process.

        Returns:
            The associated :class:`~BalancerV2Pool`.

        Examples:
            >>> pool = events._process_event(event)
        """
        try:
            specialization = PoolSpecialization(event["specialization"])
        except ValueError:
            specialization = None
        pool = BalancerV2Pool(
            address=event["poolAddress"],
            id=HexBytes(event["poolId"]),
            specialization=specialization,
            vault=self.vault,
            _deploy_block=event.block_number,
            asynchronous=self.asynchronous,
        )
        # lets get this cached into memory now
        task = create_task(pool.tokens(sync=False))
        self.__tasks.append(task)
        task.add_done_callback(self._task_done_callback)
        return pool

    def _get_block_for_obj(self, pool: "BalancerV2Pool") -> int:
        """
        Get the block number for a specific pool.

        Args:
            pool: The :class:`~BalancerV2Pool` to evaluate.

        Returns:
            The block number.

        Examples:
            >>> block_number = events._get_block_for_obj(pool)
        """
        return pool._deploy_block

    def _task_done_callback(self, t: Task):
        """
        Callback function for when a task is completed.

        Args:
            t: The completed task.

        Examples:
            >>> events._task_done_callback(task)
        """
        self.__tasks.remove(t)
        if not t.cancelled():
            # get the exc so it doesn't log, it will come up later
            t.exception()


class BalancerV2Pool(BalancerPool):
    """A pool from Balancer Protocol v2"""

    # internal variables to save calls in some instances
    # they do not necessarily reflect real life at all times
    # defaults are stored as class vars to keep instance dicts smaller
    _tokens: Tuple[ERC20, ...] = None
    __nonweighted: bool = False
    __weights: List[int] = None

    def __init__(
        self,
        address: AnyAddressType,
        *,
        id: Optional[HexBytes] = None,
        specialization: Optional[PoolSpecialization] = None,
        vault: Optional[BalancerV2Vault] = None,
        asynchronous: bool = False,
        _deploy_block: Optional[Block] = None,
    ):
        """
        Initialize a BalancerV2Pool instance.

        Args:
            address: The address of the pool.
            id: The ID of the pool.
            specialization: The specialization of the pool.
            vault: The associated :class:`~BalancerV2Vault`.
            asynchronous: Whether to use asynchronous operations.
            _deploy_block: The block number when the pool was deployed.

        Examples:
            >>> pool = BalancerV2Pool("0xPoolAddress")
        """
        super().__init__(address, asynchronous=asynchronous, _deploy_block=_deploy_block)
        if id is not None:
            self.id = id
        if specialization is not None:
            self.pool_type = specialization
        if vault is not None:
            self.vault = vault

    @a_sync.aka.cached_property
    async def id(self) -> PoolId:
        """
        Get the ID of the pool.

        Returns:
            The pool ID.

        Examples:
            >>> pool_id = await pool.id
        """
        return await Call(self.address, "getPoolId()(bytes32)")

    __id__: HiddenMethodDescriptor[Self, PoolId]

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def vault(self) -> Optional[BalancerV2Vault]:
        """
        Get the associated Balancer V2 Vault.

        Returns:
            The associated :class:`~BalancerV2Vault`, or None if not found.

        Examples:
            >>> vault = await pool.vault
        """
        with suppress(ContractLogicError):
            vault = await Call(self.address, "getVault()(address)")
            if vault == ZERO_ADDRESS:
                return None
            elif vault:
                return BalancerV2Vault(vault, asynchronous=True)
        # in earlier web3 versions, we would get `None`. More recently, we get ContractLogicError. This handles both
        if CONNECTED_TO_MAINNET and await self.__build_name__ == "CronV1Pool":
            # NOTE: these `CronV1Pool` tokens ARE balancer pools but don't match the expected pool abi?
            return BalancerV2Vault("0xBA12222222228d8Ba445958a75a0704d566BF2C8", asynchronous=True)

    __vault__: HiddenMethodDescriptor[Self, Optional[BalancerV2Vault]]

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pool_type(self) -> Union[PoolSpecialization, int]:
        """
        Get the type of the pool.

        Returns:
            The pool type as a :class:`~PoolSpecialization` or an integer.

        Examples:
            >>> pool_type = await pool.pool_type
        """
        vault = await self.__vault__
        if vault is None:
            raise ValueError(f"{self} has no vault") from None
        elif poolid := await self.__id__:
            _, specialization = await vault.contract.getPool.coroutine(poolid)
        elif CONNECTED_TO_MAINNET and await self.__build_name__ == "CronV1Pool":
            # NOTE: these `CronV1Pool` tokens ARE balancer pools but don't match the expected pool abi?
            return PoolSpecialization.CronV1Pool
        else:
            raise ValueError(f"{self} has no poolid") from None
        try:
            return PoolSpecialization(specialization)
        except ValueError:
            if self.address not in _warned:
                with suppress(ContractNotVerified):
                    logger.warning(
                        "ypricemagic does not recognize this pool type, please add `%s = %s` to %s.PoolSpecialization (pool=%s)",
                        await self.__build_name__,
                        specialization,
                        __name__,
                        self.address,
                    )
                _warned.add(self.address)
            return specialization

    __pool_type__: HiddenMethodDescriptor[Self, Optional[PoolSpecialization]]

    @stuck_coro_debugger
    async def get_tvl(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Optional[UsdValue]:
        """
        Get the total value locked (TVL) in the pool in USD.

        Args:
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            The TVL in USD, or None if it cannot be determined.

        Examples:
            >>> tvl = await pool.get_tvl()
        """
        if balances := await self.get_balances(block=block, skip_cache=skip_cache, sync=False):
            # overwrite ref to big obj with ref to little obj
            balances = iter(tuple(balances.values()))
            return UsdValue(await WeiBalance.value_usd.sum(balances, sync=False))

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_balances(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Dict[ERC20, WeiBalance]:
        """
        Get the balances of tokens in the pool.

        Args:
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            A dictionary mapping :class:`~ERC20` tokens to their :class:`~WeiBalance` in the pool.

        Examples:
            >>> balances = await pool.get_balances()
        """
        vault = await self.__vault__
        if vault is None:
            return {}
        tokens, balances, lastChangedBlock = await vault.get_pool_tokens(
            await self.__id__, block=block, sync=False
        )
        return {
            ERC20(token, asynchronous=self.asynchronous): WeiBalance(
                balance, token, block=block, skip_cache=skip_cache
            )
            for token, balance in zip(tokens, balances)
            # NOTE: some pools include themselves in their own token list, and we should ignore those
            if token != self.address
        }

    async def get_balance(
        self,
        token_address: Address,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[WeiBalance]:
        """
        Get the balance of a specific token in the pool.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            The :class:`~WeiBalance` of the specified token in the pool, or None if not found.

        Examples:
            >>> balance = await pool.get_balance("0xTokenAddress")
        """
        if info := await self.get_balances(block=block, sync=False):
            try:
                return info[token_address]
            except KeyError:
                raise TokenNotFound(token_address, self) from None

    @stuck_coro_debugger
    async def get_token_price(
        self,
        token_address: AnyAddressType,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        """
        Get the price of a specific token in the pool in USD.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            The price of the specified token in USD, or None if it cannot be determined.

        Examples:
            >>> price = await pool.get_token_price("0xTokenAddress")
        """
        get_balances_coro = self.get_balances(block=block, skip_cache=skip_cache, sync=False)
        if self.__nonweighted:
            # this await will return immediately once cached
            token_balances = await get_balances_coro
            weights = self.__weights
        else:
            token_balances, weights = await cgather(
                get_balances_coro, self.weights(block=block, sync=False)
            )
        pool_token_info = list(zip(token_balances.keys(), token_balances.values(), weights))
        for pool_token, token_balance, token_weight in pool_token_info:
            if pool_token == token_address:
                break

        paired_token_balance: Optional[WeiBalance] = None
        for pool_token, balance, weight in pool_token_info:
            if pool_token in constants.STABLECOINS:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif pool_token == constants.WRAPPED_GAS_COIN:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif len(pool_token_info) == 2 and pool_token != token_address:
                paired_token_balance, paired_token_weight = balance, weight
                break

        if paired_token_balance is None:
            return None

        token_value_in_pool, token_balance_readable = await cgather(
            paired_token_balance.__value_usd__, token_balance.__readable__
        )
        token_value_in_pool /= paired_token_weight * token_weight
        return UsdPrice(token_value_in_pool / token_balance_readable)

    # NOTE: We can't cache this as a cached property because some balancer pool tokens can change. Womp
    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def tokens(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Tuple[ERC20, ...]:
        """
        Get the tokens in the pool.

        Args:
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            A tuple of :class:`~ERC20` tokens in the pool.

        Examples:
            >>> tokens = await pool.tokens()
        """
        if self._tokens:
            return self._tokens
        tokens = tuple(
            (await self.get_balances(block=block, skip_cache=skip_cache, sync=False)).keys()
        )
        if await self.__pool_type__ in PoolSpecialization.with_immutable_tokens():
            self._tokens = tokens
        return tokens

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def weights(self, block: Optional[Block] = None) -> List[int]:
        """
        Get the weights of tokens in the pool.

        Args:
            block: The block number to query. Defaults to the latest block.

        Returns:
            A list of weights for the tokens in the pool.

        Examples:
            >>> weights = await pool.weights()
        """
        contract = await Contract.coroutine(self.address)
        try:
            return await contract.getNormalizedWeights.coroutine(block_identifier=block)
        except AttributeError:
            # Contract has no method `getNormalizedWeights`
            self.__nonweighted = True
            num_tokens = len(await self.tokens(block=block, sync=False))
            self.__weights = [10**18 // num_tokens] * num_tokens
            return self.__weights


class BalancerV2(BalancerABC[BalancerV2Pool]):

    _pool_type = BalancerV2Pool

    _check_methods = (
        "getPoolId()(bytes32)",
        "getPausedState()((bool,uint,uint))",
        "getSwapFeePercentage()(uint)",
    )

    def __init__(self, *, asynchronous: bool = False) -> None:
        """
        Initialize a BalancerV2 instance.

        Args:
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> balancer = BalancerV2(asynchronous=True)
        """
        super().__init__()
        self.asynchronous = asynchronous
        self.vaults = [
            BalancerV2Vault(vault, asynchronous=self.asynchronous) for vault in BALANCER_V2_VAULTS
        ]

    @stuck_coro_debugger
    async def get_token_price(
        self,
        token_address: Address,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> UsdPrice:
        """
        Get the price of a specific token in USD.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            The price of the specified token in USD.

        Examples:
            >>> price = await balancer.get_token_price("0xTokenAddress")
        """
        if deepest_pool := await self.deepest_pool_for(token_address, block=block, sync=False):
            return await deepest_pool.get_token_price(
                token_address, block, skip_cache=skip_cache, sync=False
            )

    # NOTE: we need a tiny semaphore here because balancer is super arduous and every unpricable token must pass thru this section
    @a_sync.Semaphore(10)
    @stuck_coro_debugger
    async def deepest_pool_for(
        self, token_address: Address, block: Optional[Block] = None
    ) -> Optional[BalancerV2Pool]:
        """
        Find the deepest pool for a specific token.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.

        Returns:
            The :class:`~BalancerV2Pool` with the deepest liquidity for the specified token, or None if not found.

        Examples:
            >>> deepest_pool = await balancer.deepest_pool_for("0xTokenAddress")
        """
        kwargs = {"token_address": token_address, "block": block}
        deepest_pools = BalancerV2Vault.deepest_pool_for.map(self.vaults, **kwargs)
        if deepest_pools := {
            vault.address: deepest_pool
            async for vault, deepest_pool in deepest_pools
            if deepest_pool is not None
        }:
            logger.debug(
                "%s deepest pools for %s at %s: %s",
                self,
                token_address,
                block,
                deepest_pools,
            )
            async for pool in (
                BalancerV2Pool.get_balance.map(deepest_pools.values(), **kwargs)
                .keys(pop=True)
                .aiterbyvalues(reverse=True)
            ):
                return pool

        # TODO: afilter
        # deepest_pools = BalancerV2Vault.deepest_pool_for.map(self.vaults, **kwargs).values(pop=True).afilter()
        # async for pool in BalancerV2Pool.get_balance.map(deepest_pools, **kwargs).keys(pop=True).aiterbyvalues(reverse=True):
        #     return pool


balancer = BalancerV2(asynchronous=True)

_lookup_balance_from_tuple: Callable[[Tuple[Any, T]], T] = (
    lambda pool_and_balance: pool_and_balance[1]
)
"Takes a tuple[K, V] and returns V."

_warned = set()



import logging
from typing import List, Optional

from a_sync import a_sync, cgather
from web3.exceptions import ContractLogicError
from web3.middleware.validation import is_not_null

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20
from y.constants import CONNECTED_TO_MAINNET
from y.contracts import has_method, has_methods
from y.datatypes import (
    Address,
    AddressOrContract,
    AnyAddressType,
    Block,
    UsdPrice,
    UsdValue,
)
from y.prices import magic
from y.utils.multicall import multicall_same_func_same_contract_different_inputs

logger = logging.getLogger(__name__)

_GET_TOKEN_INPUTS = tuple(range(8))


@a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
async def is_saddle_lp(token_address: AnyAddressType) -> bool:
    """
    Determine if a given token is a Saddle LP token.

    This function checks if the token has specific methods that are characteristic of Saddle LP tokens.

    Args:
        token_address: The address of the token to check.

    Returns:
        True if the token is a Saddle LP token, False otherwise.

    Examples:
        >>> await is_saddle_lp("0x1234567890abcdef1234567890abcdef12345678")
        False

    See Also:
        - :func:`get_pool`
    """
    pool = await get_pool(token_address, sync=False)
    return pool is not None and await has_methods(
        pool,
        ("getVirtualPrice()(uint)", "getA()(uint)", "getAPrecise()(uint)"),
        sync=False,
    )


@a_sync(default="sync", ram_cache_ttl=ENVS.CACHE_TTL)
async def get_pool(token_address: AnyAddressType) -> Address:
    """
    Retrieve the pool address for a given Saddle LP token.

    This function checks for specific token addresses on the Ethereum Mainnet and returns hardcoded pool addresses for those tokens.
    If the token is not one of the known tokens, it attempts to find a pool by checking if the token has a `swap()` method.

    Args:
        token_address: The address of the token to find the pool for.

    Returns:
        The address of the pool if found, otherwise None.

    Examples:
        >>> await get_pool("0xc9da65931ABf0Ed1b74Ce5ad8c041C4220940368")
        '0xa6018520EAACC06C30fF2e1B3ee2c7c22e64196a'

    See Also:
        - :func:`is_saddle_lp`
    """
    token_address = await convert.to_address_async(token_address)
    if CONNECTED_TO_MAINNET:
        # saddle aleth doesn't have swap() function
        if token_address == "0xc9da65931ABf0Ed1b74Ce5ad8c041C4220940368":
            return "0xa6018520EAACC06C30fF2e1B3ee2c7c22e64196a"
        elif token_address == "0xd48cF4D7FB0824CC8bAe055dF3092584d0a1726A":  # saddle d4
            return "0xC69DDcd4DFeF25D8a793241834d4cc4b3668EAD6"
        elif token_address == "0xF32E91464ca18fc156aB97a697D6f8ae66Cd21a3":
            return "0xdf3309771d2BF82cb2B6C56F9f5365C8bD97c4f2"
    try:
        pool = await has_method(token_address, "swap()(address)", return_response=True, sync=False)
    except ContractLogicError:
        return None
    else:
        return pool or None


@a_sync(default="sync")
async def get_price(token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
    """
    Calculate the price of a Saddle LP token in USD.

    This function calculates the price by dividing the total value locked (TVL) by the total supply of the token.

    Args:
        token_address: The address of the token to get the price for.
        block: The block number to query. Defaults to the latest block.

    Examples:
        >>> await get_price("0x1234567890abcdef1234567890abcdef12345678")
        UsdPrice(1.2345)

    See Also:
        - :func:`get_tvl`
    """
    tvl, total_supply = await cgather(
        get_tvl(token_address, block, sync=False),
        ERC20(token_address, asynchronous=True).total_supply_readable(block),
    )
    return UsdPrice(tvl / total_supply)


@a_sync(default="sync")
async def get_tvl(
    token_address: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdValue:
    """
    Calculate the total value locked (TVL) in a Saddle LP token pool.

    This function retrieves the pool and tokens concurrently, then uses multicall to get the balances.
    It calculates the TVL by summing the value of each token in the pool.

    Args:
        token_address: The address of the token to get the TVL for.
        block: The block number to query. Defaults to the latest block.
        skip_cache: If True, skip using the cache while fetching price data.

    Examples:
        >>> await get_tvl("0x1234567890abcdef1234567890abcdef12345678")
        UsdValue(123456.78)

    See Also:
        - :func:`get_price`
    """
    tokens: List[ERC20]
    pool, tokens = await cgather(
        get_pool(token_address, sync=False),
        get_tokens(token_address, block, sync=False),
    )
    balances = await multicall_same_func_same_contract_different_inputs(
        pool,
        "getTokenBalance(uint8)(uint)",
        inputs=range(len(tokens)),
        sync=False,
    )
    scales, prices = await cgather(
        ERC20.scale.map(tokens).values(pop=True),
        magic.get_prices(tokens, block, skip_cache=skip_cache, silent=True, sync=False),
    )
    return UsdValue(
        sum(balance / scale * price for balance, scale, price in zip(balances, scales, prices))
    )


@a_sync(default="sync")
async def get_tokens(token_address: AnyAddressType, block: Optional[Block] = None) -> List[ERC20]:
    """
    Retrieve the list of tokens in a Saddle LP token pool.

    This function retrieves the pool and then uses multicall to get the tokens in the pool.

    Args:
        token_address: The address of the token to get the tokens for.
        block: The block number to query. Defaults to the latest block.

    Examples:
        >>> await get_tokens("0x1234567890abcdef1234567890abcdef12345678")
        [ERC20('0xToken1'), ERC20('0xToken2')]

    See Also:
        - :func:`get_pool`
    """
    pool, tokens = await cgather(
        get_pool(token_address, sync=False),
        multicall_same_func_same_contract_different_inputs(
            pool,
            "getToken(uint8)(address)",
            inputs=_GET_TOKEN_INPUTS,
            block=block,
            return_None_on_failure=True,
            sync=False,
        ),
    )
    return list(map(ERC20, filter(is_not_null, tokens)))



from asyncio import iscoroutinefunction
from functools import partial, wraps
from logging import getLogger
from typing import Callable, TypeVar, Union

from a_sync import debugging
from typing_extensions import ParamSpec


P = ParamSpec("P")
T = TypeVar("T")


stuck_coro_logger = getLogger("y.stuck?")
stuck_coro_debugger = partial(debugging.stuck_coro_debugger, logger=stuck_coro_logger)


def continue_on_revert(func: Callable[P, T]) -> Callable[P, T]:
    """
    Decorates a call-making function. If the call reverts, it attempts to continue
    by calling the standalone function :func:`continue_if_call_reverted` from the
    `y.exceptions` module. Raises any other exceptions.

    This decorator is useful for functions interacting with smart contracts where
    a call might revert due to various reasons. It ensures that the function does not
    fail completely on a revert, but instead handles it gracefully.

    Note:
        The decorated function will return `None` if the call reverts and
        :func:`continue_if_call_reverted` does not raise an exception.

    Args:
        func: The function to be decorated.

    Returns:
        The wrapped function with revert handling.

    Examples:
        >>> @continue_on_revert
        ... def fetch_data():
        ...     # some logic that might revert
        ...     pass

        >>> @continue_on_revert
        ... async def fetch_data_async():
        ...     # some async logic that might revert
        ...     pass

    See Also:
        - :func:`continue_if_call_reverted`
    """
    from y.exceptions import continue_if_call_reverted

    if iscoroutinefunction(func):

        @wraps(func)
        async def continue_on_revert_wrap(*args: P.args, **kwargs: P.kwargs) -> Union[T, None]:
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                continue_if_call_reverted(e)

    elif callable(func):

        @wraps(func)
        def continue_on_revert_wrap(*args: P.args, **kwargs: P.kwargs) -> Union[T, None]:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                continue_if_call_reverted(e)

    else:
        raise NotImplementedError(f"Unable to decorate {func}")
    return continue_on_revert_wrap



import logging
from typing import List, Optional, Union

import a_sync
from a_sync.a_sync.property import HiddenMethodDescriptor
from typing_extensions import Self

from y import ENVIRONMENT_VARIABLES as ENVS
from y import exceptions
from y._decorators import stuck_coro_debugger
from y.constants import CHAINID, CONNECTED_TO_MAINNET
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices.dex.balancer._abc import BalancerABC
from y.prices.dex.balancer.v1 import BalancerV1
from y.prices.dex.balancer.v2 import BalancerV2
from y.utils.cache import optional_async_diskcache


logger = logging.getLogger(__name__)


class BalancerMultiplexer(a_sync.ASyncGenericBase):
    """A multiplexer for interacting with different versions of Balancer pools.

    This class provides methods to determine if a token is a Balancer pool,
    retrieve pool prices, and get token prices across different Balancer versions.

    Examples:
        Initialize the multiplexer:

        >>> multiplexer = BalancerMultiplexer(asynchronous=True)

        Check if a token is a Balancer pool:

        >>> is_pool = await multiplexer.is_balancer_pool(token_address)

        Get the price of a token:

        >>> price = await multiplexer.get_price(token_address, block=12345678)

    See Also:
        - :class:`BalancerV1`
        - :class:`BalancerV2`
    """

    def __init__(self, *, asynchronous: bool = False) -> None:
        """
        Initialize the BalancerMultiplexer.

        Args:
            asynchronous: Whether to operate in asynchronous mode.
        """
        super().__init__()
        self.asynchronous = asynchronous

    @a_sync.aka.property
    async def versions(self) -> List[Union[BalancerV1, BalancerV2]]:
        """
        Get the available Balancer versions.

        Returns:
            A list of available Balancer versions.

        Examples:
            >>> versions = await multiplexer.versions
        """
        return [v async for v in a_sync.as_completed([self.__v1__, self.__v2__], aiter=True) if v]

    __versions__: HiddenMethodDescriptor[Self, List[Union[BalancerV1, BalancerV2]]]

    @a_sync.aka.cached_property
    async def v1(self) -> Optional[BalancerV1]:
        """
        Get the Balancer V1 instance.

        Returns:
            An instance of BalancerV1 if available, otherwise None.

        Examples:
            >>> v1 = await multiplexer.v1
        """
        try:
            return BalancerV1(asynchronous=self.asynchronous)
        except ImportError:
            return None

    __v1__: HiddenMethodDescriptor[Self, Optional[BalancerV1]]

    @a_sync.aka.cached_property
    async def v2(self) -> Optional[BalancerV2]:
        """
        Get the Balancer V2 instance.

        Returns:
            An instance of BalancerV2 if available, otherwise None.

        Examples:
            >>> v2 = await multiplexer.v2
        """
        try:
            return BalancerV2(asynchronous=self.asynchronous)
        except ImportError:
            return None

    __v2__: HiddenMethodDescriptor[Self, Optional[BalancerV2]]

    @stuck_coro_debugger
    @optional_async_diskcache
    async def is_balancer_pool(self, token_address: AnyAddressType) -> bool:
        """
        Check if a given token address is a Balancer pool.

        Args:
            token_address: The address of the token to check.

        Returns:
            True if the token is a Balancer pool, otherwise False.

        Examples:
            >>> is_pool = await multiplexer.is_balancer_pool(token_address)
        """
        try:
            await self.get_version(token_address)
            return True
        except exceptions.TokenError:
            return False

    @stuck_coro_debugger
    async def get_pool_price(
        self,
        token_address: AnyAddressType,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        """
        Get the price of a Balancer pool.

        Args:
            token_address: The address of the pool token.
            block: The block number to query the price at.
            skip_cache: Whether to skip the cache.

        Returns:
            The price of the pool in USD, or None if not available.

        Examples:
            >>> price = await multiplexer.get_pool_price(token_address, block=12345678)
        """
        balancer: BalancerABC = await self.get_version(token_address)
        logger.debug("pool %s is from %s", token_address, balancer)
        price = await balancer.get_pool_price(
            token_address, block, skip_cache=skip_cache, sync=False
        )
        return None if price is None else UsdPrice(price)

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_price(
        self,
        token_address: AnyAddressType,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        """
        Get the price of a token using Balancer pools.

        Args:
            token_address: The address of the token.
            block: The block number to query the price at.
            skip_cache: Whether to skip the cache.

        Returns:
            The price of the token in USD, or None if not available.

        Examples:
            >>> price = await multiplexer.get_price(token_address, block=12345678)
        """
        if await self.is_balancer_pool(token_address, sync=False):
            return await self.get_pool_price(
                token_address, block=block, skip_cache=skip_cache, sync=False
            )

        price = None

        if (  # NOTE: Only query v2 if block queried > v2 deploy block plus some extra blocks to build up liquidity
            CONNECTED_TO_MAINNET and (not block or block > 12272146 + 100000)
        ) or (
            CHAINID == Network.Fantom and (not block or block > 16896080)
        ):  # TODO: refactor this out
            v2 = await self.__v2__
            if price := await v2.get_token_price(
                token_address, block, skip_cache=skip_cache, sync=False
            ):
                logger.debug("balancer v2 -> $%s", price)
                return price

        if not price and CONNECTED_TO_MAINNET:
            v1 = await self.__v1__
            if price := await v1.get_token_price(
                token_address, block, skip_cache=skip_cache, sync=False
            ):
                logger.debug("balancer v1 -> $%s", price)
                return price

    # cached forever because not many items
    @a_sync.a_sync(cache_type="memory", ram_cache_ttl=None)
    async def get_version(self, token_address: AnyAddressType) -> BalancerABC:
        """
        Determine the Balancer version for a given token address.

        Args:
            token_address: The address of the token.

        Returns:
            The Balancer version instance.

        Raises:
            exceptions.TokenError: If the token is not a Balancer pool.

        Examples:
            >>> version = await multiplexer.get_version(token_address)
        """
        for v in await self.__versions__:
            if await v.is_pool(token_address, sync=False):
                return v
        raise exceptions.TokenError(token_address, "Balancer pool")


balancer_multiplexer = BalancerMultiplexer(asynchronous=True)



import os
from contextlib import suppress

from brownie import chain, network

try:
    # leaving this in here for backward compatability with older brownie versions
    # TODO: remove me when eetherscan disables v1 api
    from brownie.network.contract import _explorer_tokens
except ImportError:
    pass
else:
    # this helps ensure backwards compatability with older versions of brownie
    if "base" not in _explorer_tokens:
        _explorer_tokens["base"] = "BASESCAN_TOKEN"
    if "optimistic" not in _explorer_tokens:
        _explorer_tokens["optimistic"] = "OPTIMISMSCAN_TOKEN"


class NetworkNotSpecified(Exception):
    pass


if not network.is_connected():
    if "BROWNIE_NETWORK_ID" not in os.environ:
        raise NetworkNotSpecified(
            "In order to use pricemagic outside of a brownie project directory, you will need to set $BROWNIE_NETWORK_ID environment variable with the id of your preferred brownie network connection."
        )
    network.connect(os.environ["BROWNIE_NETWORK_ID"])

# for backwards-compatability
from dank_mids import dank_web3 as dank_w3


from y import convert, time
from y.classes.common import ERC20
from y.constants import EEE_ADDRESS, WRAPPED_GAS_COIN, dai, usdc, wbtc, weth
from y.contracts import (
    Contract,
    Contract_erc20,
    Contract_with_erc20_fallback,
    contract_creation_block,
    contract_creation_block_async,
    has_method,
    has_methods,
)
from y.datatypes import Address, Block
from y.exceptions import (
    CalldataPreparationError,
    CallReverted,
    ContractNotVerified,
    MessedUpBrownieContract,
    NonStandardERC20,
    NotABalancerV2Pool,
    NotAUniswapV2Pool,
    PriceError,
    UnsupportedNetwork,
)
from y import exceptions, monkey_patches
from y.networks import Network
from y.prices import magic
from y.prices.magic import get_price, get_prices, map_prices
from y.prices.utils import check_bucket
from y.time import (
    get_block_at_timestamp,
    get_block_timestamp,
    get_block_timestamp_async,
)
from y.utils.events import Events, LogFilter, ProcessedEvents
from y.utils.logging import enable_debug_logging
from y.utils.middleware import setup_getcode_cache_middleware, setup_geth_poa_middleware
from y.utils.multicall import fetch_multicall
from y.utils.raw_calls import balanceOf, raw_call

setup_getcode_cache_middleware()

monkey_patches.monkey_patch_checksum_cache()

if chain.id == Network.Optimism:
    setup_geth_poa_middleware()

__all__ = [
    ### you can reach the below functions, classes, and variables using ###
    ###  `y.__name__` or `from y import __name__`.  ###
    # prices
    "get_price",
    "get_prices",
    "map_prices",
    "check_bucket",
    # erc20
    "ERC20",
    # logs filter,
    "LogFilter",
    # decoded events filter
    "Events",
    "ProcessedEvents",
    # contract stuff
    "Contract",
    "contract_creation_block",
    "contract_creation_block_async",
    "has_method",
    "has_methods",
    # network stuff
    "Network",
    # constants
    "EEE_ADDRESS",
    "WRAPPED_GAS_COIN",
    "weth",
    "dai",
    "usdc",
    "wbtc",
    # magic
    "magic",
    # raw calls
    "raw_call",
    # convert
    "convert",
    # time
    "time",
    "get_block_at_timestamp",
    "get_block_timestamp",
    "get_block_timestamp_async",
    # monkey patches,
    "monkey_patches",
]

with suppress(ModuleNotFoundError):
    """If eth_portfolio is also installed in this env, we will use its extended version of our db schema"""
    from eth_portfolio._db import entities as _db_entities



ERC20ABI = [
    {
        "inputs": [{"internalType": "uint256", "name": "chainId_", "type": "uint256"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "src",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "guy",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "wad",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": True,
        "inputs": [
            {
                "indexed": True,
                "internalType": "bytes4",
                "name": "sig",
                "type": "bytes4",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "usr",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "bytes32",
                "name": "arg1",
                "type": "bytes32",
            },
            {
                "indexed": True,
                "internalType": "bytes32",
                "name": "arg2",
                "type": "bytes32",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "data",
                "type": "bytes",
            },
        ],
        "name": "LogNote",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "src",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "dst",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "wad",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "DOMAIN_SEPARATOR",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "PERMIT_TYPEHASH",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "address", "name": "", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "usr", "type": "address"},
            {"internalType": "uint256", "name": "wad", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "usr", "type": "address"},
            {"internalType": "uint256", "name": "wad", "type": "uint256"},
        ],
        "name": "burn",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "address", "name": "guy", "type": "address"}],
        "name": "deny",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "usr", "type": "address"},
            {"internalType": "uint256", "name": "wad", "type": "uint256"},
        ],
        "name": "mint",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "src", "type": "address"},
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "wad", "type": "uint256"},
        ],
        "name": "move",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "", "type": "address"}],
        "name": "nonces",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "holder", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "nonce", "type": "uint256"},
            {"internalType": "uint256", "name": "expiry", "type": "uint256"},
            {"internalType": "bool", "name": "allowed", "type": "bool"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "permit",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "usr", "type": "address"},
            {"internalType": "uint256", "name": "wad", "type": "uint256"},
        ],
        "name": "pull",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "usr", "type": "address"},
            {"internalType": "uint256", "name": "wad", "type": "uint256"},
        ],
        "name": "push",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [{"internalType": "address", "name": "guy", "type": "address"}],
        "name": "rely",
        "outputs": [],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "wad", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": False,
        "inputs": [
            {"internalType": "address", "name": "src", "type": "address"},
            {"internalType": "address", "name": "dst", "type": "address"},
            {"internalType": "uint256", "name": "wad", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "payable": False,
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [],
        "name": "version",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
    {
        "constant": True,
        "inputs": [{"internalType": "address", "name": "", "type": "address"}],
        "name": "wards",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "payable": False,
        "stateMutability": "view",
        "type": "function",
    },
]
